"""Liquid exception test cases."""
# pylint: disable=missing-class-docstring

from pathlib import Path
from unittest import TestCase

from liquid import Environment
from liquid import Mode

from liquid.exceptions import Error
from liquid.exceptions import LiquidSyntaxError


class EnvironmentErrorTestCase(TestCase):
    def test_exception_class_is_raised(self):
        """Test env.error handles exception classes."""
        env = Environment(tolerance=Mode.STRICT, strict_filters=True)

        with self.assertRaises(Error):
            env.error(LiquidSyntaxError, msg=":(")


class LiquidErrorTestCase(TestCase):
    def test_base_error_message(self):
        """Test that the base error can include a message."""
        try:
            raise Error("Oh no!", "extra info")
        except Error as err:
            self.assertEqual(err.message, "Oh no!")

    def test_base_error_no_message(self):
        """Test that the base error can not include a message."""
        try:
            raise Error()
        except Error as err:
            self.assertEqual(err.message, None)

    def test_base_error_include_filename(self):
        """Test that the base error can include a filename."""
        try:
            raise Error("Oh no!", linenum=1, filename="foo.liquid")
        except Error as err:
            self.assertEqual(str(err), "Oh no!, on line 1 of foo.liquid")


class LiquidSyntaxErrorTestCase(TestCase):
    def test_template_name_from_string(self):
        """Test that a syntax error can include a template name as a string."""
        try:
            raise LiquidSyntaxError("Oh no!", filename="foo.liquid")
        except LiquidSyntaxError as err:
            self.assertEqual(err.name, "foo.liquid")

    def test_template_name_from_path(self):
        """Test that a syntax error can include a template name as a path."""
        try:
            raise LiquidSyntaxError("Oh no!", filename=Path("/templates/foo.liquid"))
        except LiquidSyntaxError as err:
            self.assertEqual(err.name, "/templates/foo.liquid")

    def test_no_template_name(self):
        """Test that a syntax error can not include template name."""
        try:
            raise LiquidSyntaxError("Oh no!")
        except LiquidSyntaxError as err:
            self.assertEqual(err.name, "")



"""Golden tests cases for testing liquid's built-in `capitalize` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="lower case string",
        template=r'{{ "hello" | capitalize }}',
        expect="Hello",
    ),
    Case(
        description="already capitalized string",
        template=r'{{ "Hello" | capitalize }}',
        expect="Hello",
    ),
    Case(
        description="unexpected argument",
        template=r'{{ "hello" | capitalize: 2 }}',
        expect="",
        error=True,
    ),
    Case(
        description="undefined left value",
        template=r"{{ nosuchthing | capitalize }}",
        expect="",
    ),
]



# flake8: noqa
# pylint: disable=missing-module-docstring
from .lex import tokenize
from .lex import tokenize_with_parens
from .parse import parse
from .parse import parse_with_parens

__all__ = (
    "parse",
    "parse_with_parens",
    "tokenize",
    "tokenize_with_parens",
)



"""Test that we can traverse a template's syntax tree to find comments"""
# pylint: disable=missing-class-docstring missing-function-docstring

import unittest

from typing import Iterator
from typing import Tuple

from liquid import Template
from liquid.ast import Node
from liquid.builtin.tags.comment_tag import CommentNode
from liquid.template import BoundTemplate


class FindCommentsTestCase(unittest.TestCase):
    def _find_comment_text(self, template: BoundTemplate) -> Iterator[Tuple[int, str]]:
        def visit(node: Node) -> Iterator[Tuple[int, str]]:
            if isinstance(node, CommentNode) and node.text is not None:
                yield (node.tok.linenum, node.text)
            for child in node.children():
                if child.node:
                    yield from visit(child.node)

        for child in template.tree.statements:
            yield from visit(child)

    def test_find_comment_tag_text(self):
        """Test that we can get text from comment tags."""
        template = Template(
            "\n".join(
                (
                    "{% comment %}hello{% endcomment %}",
                    "{% if false %}",
                    "{% comment %}",
                    "foo bar",
                    "{% endcomment %}",
                    "{% endif %}",
                    "{% for x in (1..3) %}",
                    "{% if true %}",
                    "{% comment %}goodbye{% endcomment %}",
                    "{% endif %}",
                    "{% endfor %}",
                    "{% comment %}world{% endcomment %}",
                )
            )
        )

        comments = list(self._find_comment_text(template))
        self.assertEqual(len(comments), 4)
        self.assertEqual(
            comments,
            [
                (1, "hello"),
                (3, "\nfoo bar\n"),
                (9, "goodbye"),
                (12, "world"),
            ],
        )

    def test_find_inline_comment_tag_text(self):
        """Test that we can get text from inline comment tags."""
        template = Template(
            "\n".join(
                (
                    "{% # hello %}",
                    "{% if false %}",
                    "{% #",
                    "# foo bar",
                    "# foo bar",
                    "%}",
                    "{% endif %}",
                    "{% for x in (1..3) %}",
                    "{% if true %}",
                    "{% # goodbye %}",
                    "{% endif %}",
                    "{% endfor %}",
                    "{% # world %}",
                )
            )
        )

        comments = list(self._find_comment_text(template))
        self.assertEqual(len(comments), 4)
        self.assertEqual(
            comments,
            [
                (1, "hello"),
                (3, "# foo bar\n# foo bar"),
                (10, "goodbye"),
                (13, "world"),
            ],
        )



"""Golden tests cases for testing liquid's built-in `times` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="int times int",
        template=r"{{ 5 | times: 2 }}",
        expect="10",
    ),
    Case(
        description="int times float",
        template=r"{{ 5 | times: 2.1 }}",
        expect="10.5",
    ),
    Case(
        description="float times float",
        template=r"{{ 5.0 | times: 2.1 }}",
        expect="10.5",
    ),
    Case(
        description="string times string",
        template=r'{{ "5.0" | times: "2.1" }}',
        expect="10.5",
    ),
    Case(
        description="negative multiplication",
        template=r"{{ -5 | times: 2 }}",
        expect="-10",
    ),
    Case(
        description="missing arg",
        template=r"{{ 5 | times }}",
        expect="",
        error=True,
    ),
    Case(
        description="too many args",
        template=r"{{ 5 | times: 1, 2 }}",
        expect="",
        error=True,
    ),
    Case(
        description="undefined left value",
        template=r"{{ nosuchthing | times: 2 }}",
        expect="0",
    ),
    Case(
        description="undefined argument",
        template=r"{{ 5 | times: nosuchthing }}",
        expect="0",
    ),
]



"""Parse tree node and pseudo "tag" for output statements."""
from typing import Any
from typing import List
from typing import Optional
from typing import TextIO

try:
    from markupsafe import escape
    from markupsafe import Markup
    from markupsafe import soft_str
except ImportError:
    from liquid.exceptions import escape  # type: ignore
    from liquid.exceptions import Markup  # type: ignore

    # pylint: disable=invalid-name
    soft_str = str  # type: ignore

from liquid.ast import ChildNode
from liquid.ast import Node

from liquid.context import Context
from liquid.expression import Expression
from liquid.parse import expect

from liquid.stream import TokenStream
from liquid.tag import Tag

from liquid.token import Token
from liquid.token import TOKEN_STATEMENT


class StatementNode(Node):
    """Parse tree node representing an output statement."""

    __slots__ = ("tok", "expression")

    force_output = True

    def __init__(self, tok: Token, expression: Expression):
        self.tok = tok
        self.expression = expression

    def __str__(self) -> str:
        return f"`{self.expression}`"

    def __repr__(self) -> str:  # pragma: no cover
        return f"StatementNode(tok={self.tok}, expression={self.expression!r})"

    def _to_liquid_string(self, val: Any, autoescape: bool) -> str:
        if isinstance(val, str) or (autoescape and hasattr(val, "__html__")):
            pass
        elif isinstance(val, bool):
            val = str(val).lower()
        elif val is None:
            val = ""
        elif isinstance(val, list):
            if autoescape:
                val = Markup("").join(soft_str(itm) for itm in val)
            else:
                val = "".join(soft_str(itm) for itm in val)
        elif isinstance(val, range):
            val = f"{val.start}..{val.stop - 1}"
        else:
            val = str(val)

        if autoescape:
            val = escape(val)

        assert isinstance(val, str)
        return val

    # pylint: disable=useless-return
    def render_to_output(self, context: Context, buffer: TextIO) -> Optional[bool]:
        val = self.expression.evaluate(context)
        buffer.write(self._to_liquid_string(val, context.autoescape))
        return None

    async def render_to_output_async(
        self, context: Context, buffer: TextIO
    ) -> Optional[bool]:
        val = await self.expression.evaluate_async(context)
        buffer.write(self._to_liquid_string(val, context.autoescape))
        return None

    def children(self) -> List[ChildNode]:
        return [ChildNode(linenum=self.tok.linenum, expression=self.expression)]


class Statement(Tag):
    """Pseudo "tag" to register output statements with the environment."""

    name = TOKEN_STATEMENT
    block = False

    def parse(self, stream: TokenStream) -> StatementNode:
        tok = stream.current
        expect(stream, TOKEN_STATEMENT)
        return StatementNode(tok, self.env.parse_filtered_expression_value(tok.value))



"""Test that all standard filters have been registered correctly."""

import datetime
import unittest
from typing import NamedTuple

from liquid import Environment
from liquid.extra import add_filters


class Case(NamedTuple):
    """Table-driven test helper."""

    description: str
    template: str
    context: dict
    expect: str


class FilterRegisterTestCase(unittest.TestCase):
    """Test that all standard filters have been registered correctly."""

    def test_filters(self) -> None:
        """Standard filter tests."""
        test_cases = [
            Case(
                description="append",
                template=r"{{ 'hello ' | append: foo }}",
                context={"foo": "there"},
                expect="hello there",
            ),
            Case(
                description="capitalize",
                template=r"{{ 'helLO' | capitalize }}",
                context={},
                expect="Hello",
            ),
            Case(
                description="downcase",
                template=r"{{ 'helLO' | downcase }}",
                context={},
                expect="hello",
            ),
            Case(
                description="escape",
                template=r"{{ foo | escape }}",
                context={"foo": "<p>hello</p>"},
                expect="&lt;p&gt;hello&lt;/p&gt;",
            ),
            Case(
                description="escape once",
                template=r"{{ foo | escape_once }}",
                context={"foo": "&lt;p&gt;hello&lt;/p&gt;"},
                expect="&lt;p&gt;hello&lt;/p&gt;",
            ),
            Case(
                description="lstrip",
                template=r"{{ foo | lstrip }}",
                context={"foo": " \t\r\n  hello\t"},
                expect="hello\t",
            ),
            Case(
                description="newline to br",
                template=r"{{ foo | newline_to_br }}",
                context={"foo": "- apples\n- oranges\n"},
                expect="- apples<br />\n- oranges<br />\n",
            ),
            Case(
                description="prepend",
                template=r"{{ foo | prepend: bar }}",
                context={"foo": "hello", "bar": "there"},
                expect="therehello",
            ),
            Case(
                description="remove",
                template=r"{{ foo | remove: bar }}",
                context={
                    "foo": "I strained to see the train through the rain",
                    "bar": "rain",
                },
                expect="I sted to see the t through the ",
            ),
            Case(
                description="remove first",
                template=r"{{ foo | remove_first: bar }}",
                context={
                    "foo": "I strained to see the train through the rain",
                    "bar": "rain",
                },
                expect="I sted to see the train through the rain",
            ),
            Case(
                description="replace",
                template=r"{{ foo | replace: bar, baz }}",
                context={
                    "foo": "Take my protein pills and put my helmet on",
                    "bar": "my",
                    "baz": "your",
                },
                expect="Take your protein pills and put your helmet on",
            ),
            Case(
                description="replace first",
                template=r"{{ foo | replace_first: bar, baz }}",
                context={
                    "foo": "Take my protein pills and put my helmet on",
                    "bar": "my",
                    "baz": "your",
                },
                expect="Take your protein pills and put my helmet on",
            ),
            Case(
                description="slice",
                template=r"{{ foo | slice: 1, 3 }}",
                context={"foo": "hello"},
                expect="ell",
            ),
            Case(
                description="split",
                template=r"{{ foo | split: ' ' | join: '#' }}",
                context={"foo": "Hi, how are you today?"},
                expect="Hi,#how#are#you#today?",
            ),
            Case(
                description="upcase",
                template=r"{{ foo | upcase }}",
                context={"foo": "hello"},
                expect="HELLO",
            ),
            Case(
                description="strip",
                template=r"{{ foo | strip }}",
                context={"foo": " \t\r\n  hello  \t\r\n "},
                expect="hello",
            ),
            Case(
                description="rstrip",
                template=r"{{ foo | rstrip }}",
                context={"foo": " \t\r\n  hello  \t\r\n "},
                expect=" \t\r\n  hello",
            ),
            Case(
                description="strip html",
                template=r"{{ foo | strip_html }}",
                context={"foo": "Have <em>you</em> read <b>Ulysses</b> &amp; &#20;?"},
                expect="Have you read Ulysses &amp; &#20;?",
            ),
            Case(
                description="strip newlines",
                template=r"{{ foo | strip_newlines }}",
                context={"foo": "hello there\nyou"},
                expect="hello thereyou",
            ),
            Case(
                description="truncate",
                template=r"{{ foo | truncate: 20 }}",
                context={"foo": "Ground control to Major Tom."},
                expect="Ground control to...",
            ),
            Case(
                description="truncate words",
                template=r"{{ foo | truncatewords: 3 }}",
                context={"foo": "Ground control to Major Tom."},
                expect="Ground control to...",
            ),
            Case(
                description="url encode",
                template=r"{{ foo | url_encode }}",
                context={"foo": "email address is bob@example.com!"},
                expect=r"email+address+is+bob%40example.com%21",
            ),
            Case(
                description="url decode",
                template=r"{{ foo | url_decode }}",
                context={"foo": r"email+address+is+bob%40example.com%21"},
                expect="email address is bob@example.com!",
            ),
            Case(
                description="size",
                template=r"{{ foo | size }}",
                context={"foo": [1, 2, 3]},
                expect="3",
            ),
            Case(
                description="default",
                template=r"{{ foo | default: 'foo' }}",
                context={"foo": None},
                expect="foo",
            ),
            Case(
                description="date",
                template=r"{{ foo | date: '%a, %b %d, %y' }}",
                context={"foo": datetime.datetime(2002, 1, 1, 11, 45, 13)},
                expect="Tue, Jan 01, 02",
            ),
            Case(
                description="abs",
                template=r"{{ foo | abs }}",
                context={"foo": -5},
                expect="5",
            ),
            Case(
                description="at most",
                template=r"{{ foo | at_most: 5 }}",
                context={"foo": 8},
                expect="5",
            ),
            Case(
                description="at least",
                template=r"{{ foo | at_least: 8 }}",
                context={"foo": 5},
                expect="8",
            ),
            Case(
                description="ceil",
                template=r"{{ foo | ceil }}",
                context={"foo": 5.4},
                expect="6",
            ),
            Case(
                description="floor",
                template=r"{{ foo | floor }}",
                context={"foo": 5.4},
                expect="5",
            ),
            Case(
                description="divided by",
                template=r"{{ foo | divided_by: 2.0 }}",
                context={"foo": 10},
                expect="5.0",
            ),
            Case(
                description="minus",
                template=r"{{ foo | minus: 2 }}",
                context={"foo": 10},
                expect="8",
            ),
            Case(
                description="plus",
                template=r"{{ foo | plus: 2 }}",
                context={"foo": 10},
                expect="12",
            ),
            Case(
                description="round",
                template=r"{{ foo | round }}",
                context={"foo": 5.6},
                expect="6",
            ),
            Case(
                description="times",
                template=r"{{ foo | times: 2 }}",
                context={"foo": 5},
                expect="10",
            ),
            Case(
                description="modulo",
                template=r"{{ foo | modulo: 7.0 }}",
                context={"foo": 10.1},
                expect="3.1",
            ),
            Case(
                description="join",
                template=r"{{ foo | join: '#' }}",
                context={"foo": ["a", "b"]},
                expect="a#b",
            ),
            Case(
                description="first",
                template=r"{{ foo | first }}",
                context={"foo": ["a", "b"]},
                expect="a",
            ),
            Case(
                description="last",
                template=r"{{ foo | last }}",
                context={"foo": ["a", "b"]},
                expect="b",
            ),
            Case(
                description="concat",
                template=r"{{ foo | concat: bar | join: '#' }}",
                context={"foo": ["a", "b"], "bar": ["c", "d"]},
                expect="a#b#c#d",
            ),
            Case(
                description="map",
                template=r"{{ foo | map: 'title' | join: '#' }}",
                context={"foo": [{"title": "foo"}, {"title": "bar"}, {"title": "baz"}]},
                expect="foo#bar#baz",
            ),
            Case(
                description="reverse",
                template=r"{{ foo | reverse | join: '#' }}",
                context={"foo": ["b", "a", "B", "A"]},
                expect="A#B#a#b",
            ),
            Case(
                description="sort",
                template=r"{{ foo | sort | join: '#' }}",
                context={"foo": ["b", "a", "C", "B", "A"]},
                expect="A#B#C#a#b",
            ),
            Case(
                description="sort natural",
                template=r"{{ foo | sort_natural | join: '#' }}",
                context={"foo": ["b", "a", "C", "B", "A"]},
                expect="a#A#b#B#C",
            ),
            Case(
                description="where",
                template=r"{{ foo | where: 'title' }}",
                context={"foo": [{"title": "foo"}, {"title": "bar"}, {"title": None}]},
                expect="{'title': 'foo'}{'title': 'bar'}",
            ),
            Case(
                description="uniq",
                template=r"{{ foo | uniq | join: '#' }}",
                context={"foo": ["a", "b", "b", "a"]},
                expect="a#b",
            ),
            Case(
                description="compact",
                template=r"{{ foo | compact | join: '#' }}",
                context={"foo": ["b", "a", None, "A"]},
                expect="b#a#A",
            ),
            Case(
                description="base64_encode",
                template=r"{{ 'one two three' | base64_encode }}",
                context={},
                expect="b25lIHR3byB0aHJlZQ==",
            ),
            Case(
                description="base64_decode",
                template=r"{{ 'b25lIHR3byB0aHJlZQ==' | base64_decode }}",
                context={},
                expect="one two three",
            ),
            Case(
                description="base64_url_safe_encode",
                template=(
                    r"{{ "
                    r"'abcdefghijklmnopqrstuvwxyz "
                    r"ABCDEFGHIJKLMNOPQRSTUVWXYZ "
                    r"1234567890 !@#$%^&*()-=_+/?.:;[]{}\|' "
                    r"| base64_url_safe_encode }}"
                ),
                context={},
                expect=(
                    "YWJjZGVmZ2hpamtsbW5vcHFyc3R1dnd4eXogQUJDREVGR0hJSktMTU5PUFFSU1"
                    "RVVldYWVogMTIzNDU2Nzg5MCAhQCMkJV4mKigpLT1fKy8_Ljo7W117fVx8"
                ),
            ),
            Case(
                description="base64_url_safe_decode",
                template=(
                    r"{{ "
                    r"'YWJjZGVmZ2hpamtsbW5vcHFyc3R1dnd4eXogQUJDREVGR0hJSktMTU5PUFFSU1"
                    r"RVVldYWVogMTIzNDU2Nzg5MCAhQCMkJV4mKigpLT1fKy8_Ljo7W117fVx8' "
                    r"| base64_url_safe_decode }}"
                ),
                context={},
                expect=(
                    r"abcdefghijklmnopqrstuvwxyz "
                    r"ABCDEFGHIJKLMNOPQRSTUVWXYZ "
                    r"1234567890 !@#$%^&*()-=_+/?.:;[]{}\|"
                ),
            ),
        ]

        env = Environment()

        for case in test_cases:
            with self.subTest(msg=case.description):
                template = env.from_string(case.template)
                result = template.render(**case.context)
                self.assertEqual(result, case.expect)


class ExtraFilterRegistrationTestCase(unittest.TestCase):
    """Test that we can add all extra filters."""

    def test_extra_filters(self) -> None:
        """Extra filter tests."""
        test_cases = [
            Case(
                description="array index",
                template="{{ array | index: 'b' }}",
                context={"array": ["a", "b", "c"]},
                expect="1",
            ),
            Case(
                description="stylesheet tag",
                template="{{ url | stylesheet_tag }}",
                context={"url": "assets/style.css"},
                expect=(
                    '<link href="assets/style.css" rel="stylesheet" '
                    'type="text/css" media="all" />'
                ),
            ),
            Case(
                description="stylesheet tag",
                template="{{ url | script_tag }}",
                context={"url": "assets/app.js"},
                expect='<script src="assets/app.js" type="text/javascript"></script>',
            ),
            Case(
                description="JSON",
                template="{{ data | json }}",
                context={"data": {"foo": [1, 2, 3]}},
                expect='{"foo": [1, 2, 3]}',
            ),
        ]

        env = Environment()
        add_filters(env)

        for case in test_cases:
            with self.subTest(msg=case.description):
                template = env.from_string(case.template)
                result = template.render(**case.context)
                self.assertEqual(result, case.expect)



"""Tag and node definition for the built-in "decrement" tag."""

import sys

from typing import List
from typing import TextIO
from typing import Optional

from liquid import ast
from liquid.context import Context
from liquid.stream import TokenStream
from liquid.tag import Tag
from liquid.parse import expect

from liquid.token import Token
from liquid.token import TOKEN_TAG
from liquid.token import TOKEN_EXPRESSION

from liquid.expressions import TokenStream as ExprTokenStream
from liquid.expressions.common import parse_unchained_identifier
from liquid.expressions.filtered.lex import tokenize

TAG_DECREMENT = sys.intern("decrement")


class DecrementNode(ast.Node):
    """Parse tree node for the built-in "decrement" tag."""

    __slots__ = ("tok", "identifier")

    def __init__(self, tok: Token, identifier: str):
        self.tok = tok
        self.identifier = identifier

    def __str__(self) -> str:
        return f"{self.identifier} -= 1"

    def render_to_output(self, context: Context, buffer: TextIO) -> Optional[bool]:
        buffer.write(str(context.decrement(self.identifier)))
        return True

    def children(self) -> List[ast.ChildNode]:
        return [
            ast.ChildNode(
                linenum=self.tok.linenum,
                template_scope=[self.identifier],
            )
        ]


class DecrementTag(Tag):
    """The built-in "decrement" tag."""

    name = TAG_DECREMENT
    block = False

    def parse(self, stream: TokenStream) -> DecrementNode:
        expect(stream, TOKEN_TAG, value=TAG_DECREMENT)
        tok = stream.current
        stream.next_token()
        expect(stream, TOKEN_EXPRESSION)
        return DecrementNode(
            tok=tok,
            identifier=str(
                parse_unchained_identifier(
                    ExprTokenStream(tokenize(stream.current.value))
                )
            ),
        )



"""Liquid template definition."""
from __future__ import annotations

import re

from collections import abc
from collections import Counter
from collections import defaultdict

from io import StringIO
from pathlib import Path

from typing import Awaitable
from typing import DefaultDict
from typing import Dict
from typing import List
from typing import Any
from typing import Iterator
from typing import Mapping
from typing import Optional
from typing import TextIO
from typing import Tuple
from typing import Union
from typing import TYPE_CHECKING

from liquid.ast import ChildNode
from liquid.ast import Node

from liquid.context import Context
from liquid.context import VariableCaptureContext
from liquid.context import ReadOnlyChainMap

from liquid.exceptions import TemplateTraversalError, LiquidInterrupt
from liquid.exceptions import LiquidSyntaxError
from liquid.exceptions import Error
from liquid.exceptions import TemplateNotFound

from liquid.expression import Expression, StringLiteral
from liquid.expression import Identifier
from liquid.expression import Literal

from liquid.output import LimitedStringIO


if TYPE_CHECKING:  # pragma: no cover
    from liquid import Environment
    from liquid.ast import ParseTree
    from liquid.loaders import UpToDate


class BoundTemplate:
    """A liquid template that has been parsed and is bound to a
    :class:`liquid.Environment`.

    You probably don't want to instantiate :class:`BoundTemplate` directly. Use
    :meth:`liquid.Environment.from_string` or :meth:`liquid.Environment.get_template`
    instead.

    :param env: The environment this template is bound to.
    :type env: liquid.Environment
    :param parse_tree: The parse tree representing this template.
    :type parse_tree: liquid.ast.ParseTree
    :param name: Optional name of the template. Defaults to an empty string.
    :type name: Optional[str]
    :param path: Optional origin path or identifier for the template.
    :type path: Optional[Union[str, Path]]
    :param globals: An optional mapping of context variables made available every
        time the resulting template is rendered. Defaults to ``None``.
    :type globals: Optional[Dict[str, object]]
    :param matter: Optional mapping containing variables associated with the template.
        Could be "front matter" or other meta data.
    :type matter: Optional[Mapping[str, object]]
    :param uptodate: Optional callable that will return ``True`` if the template is up
        to date, or ``False`` if it needs to be reloaded. Defaults to ``None``.
    :type uptodate: Optional[Callable[[], bool]]
    """

    # Subclass `BoundTemplate` and override `context_class` to use a subclass of
    # `Context` when rendering templates.
    context_class = Context
    capture_context_class = VariableCaptureContext

    # pylint: disable=redefined-builtin, too-many-arguments
    def __init__(
        self,
        env: Environment,
        parse_tree: ParseTree,
        name: str = "",
        path: Optional[Union[str, Path]] = None,
        globals: Optional[Dict[str, object]] = None,
        matter: Optional[Mapping[str, object]] = None,
        uptodate: UpToDate = None,
    ):
        self.env = env
        self.tree = parse_tree
        self.globals = globals or {}
        self.matter = matter or {}
        self.name = name
        self.path = path
        self.uptodate = uptodate

    def render(self, *args: Any, **kwargs: Any) -> str:
        """Render the template with `args` and `kwargs` included in the render context.

        Accepts the same arguments as the :class:`dict` constructor.
        """
        context = self.context_class(
            self.env, globals=self.make_globals(dict(*args, **kwargs))
        )
        buf = self._get_buffer()
        self.render_with_context(context, buf)
        return buf.getvalue()

    async def render_async(self, *args: Any, **kwargs: Any) -> str:
        """An async version of :meth:`liquid.template.BoundTemplate.render`."""
        context = self.context_class(
            self.env, globals=self.make_globals(dict(*args, **kwargs))
        )
        buf = self._get_buffer()
        await self.render_with_context_async(context, buf)
        return buf.getvalue()

    def _get_buffer(self) -> StringIO:
        if self.env.output_stream_limit is None:
            return StringIO()
        return LimitedStringIO(limit=self.env.output_stream_limit)

    def render_with_context(
        self,
        context: Context,
        buffer: TextIO,
        *args: Any,
        partial: bool = False,
        block_scope: bool = False,
        **kwargs: Any,
    ) -> None:
        """Render the template using an existing context and output buffer.

        ``args`` and ``kwargs`` are passed to the :class:`dict` constructor. The
        resulting dictionary is added to the render context.

        :param context: A render context.
        :param buffer: File-like object to which rendered text is written.
        :param partial: If `True`, indicates that the current template has been
            included using either a "render" or "include" tag. Defaults to ``False``.
        :param block_scope: If `True`, indicates that assigns, breaks and continues
            from this template will not leak into the parent context. Defaults to
            ``False``.
        """
        # "template" could get overridden from args/kwargs, "partial" will not.
        namespace = self.make_partial_namespace(partial, dict(*args, **kwargs))

        with context.extend(namespace=namespace):
            for node in self.tree.statements:
                try:
                    node.render(context, buffer)
                except LiquidInterrupt as err:
                    # If this is an "included" template, there could be a for loop
                    # in a parent template. A for loop that could be interrupted
                    # from an included template.
                    #
                    # Convert the interrupt to a syntax error if there is no parent.
                    if not partial or block_scope:
                        self.env.error(
                            LiquidSyntaxError(
                                f"unexpected '{err}'", linenum=node.token().linenum
                            )
                        )
                    else:
                        raise
                except Error as err:
                    # Raise or warn according to the current mode.
                    self.env.error(err, linenum=node.token().linenum)

    async def render_with_context_async(
        self,
        context: Context,
        buffer: TextIO,
        *args: Any,
        partial: bool = False,
        block_scope: bool = False,
        **kwargs: Any,
    ) -> None:
        """An async version of `render_with_context`."""
        # "template" could get overridden from args/kwargs, "partial" will not.
        namespace = self.make_partial_namespace(partial, dict(*args, **kwargs))

        with context.extend(namespace=namespace):
            for node in self.tree.statements:
                try:
                    await node.render_async(context, buffer)
                except LiquidInterrupt as err:
                    # If this is an "included" template, there could be a for loop
                    # in a parent template. A for loop that could be interrupted
                    # from an included template.
                    #
                    # Convert the interrupt to a syntax error if there is no parent.
                    if not partial or block_scope:
                        self.env.error(
                            LiquidSyntaxError(
                                f"unexpected '{err}'", linenum=node.token().linenum
                            )
                        )
                    else:
                        raise
                except Error as err:
                    # Raise or warn according to the current mode.
                    self.env.error(err, linenum=node.token().linenum)

    @property
    def is_up_to_date(self) -> bool:
        """False if the template was modified since it was last parsed,
        True otherwise."""
        if not self.uptodate:
            return True

        uptodate = self.uptodate()
        if not isinstance(uptodate, bool):
            raise Error(
                f"expected a boolean from uptodate, found {type(uptodate).__name__}"
            )
        return uptodate

    async def is_up_to_date_async(self) -> bool:
        """An async version of the ``is_up_to_date`` property.

        If ``template.uptodate`` is a coroutine, it wil be awaited. Otherwise it will be
        called just like ``is_up_to_date``
        """
        if not self.uptodate:
            return True

        uptodate = self.uptodate()
        if isinstance(uptodate, Awaitable):
            return await uptodate
        return uptodate

    def make_globals(self, render_args: Mapping[str, object]) -> Mapping[str, object]:
        """Return a mapping aggregated from render arguments, template globals and
        matter variables."""
        return ReadOnlyChainMap(
            render_args,
            self.matter,
            self.globals,
        )

    def make_partial_namespace(
        self,
        partial: bool,
        render_args: Mapping[str, object],
    ) -> Mapping[str, object]:
        """Return a namespace dictionary. This is used by `render_with_context` to
        extend an existing context."""
        return {**render_args, "partial": partial}

    def __repr__(self) -> str:
        return (
            f"Template(name='{self.name}', "
            f"path='{self.path}', uptodate={self.is_up_to_date})"
        )  # pragma: no cover

    def analyze(
        self, follow_partials: bool = True, raise_for_failures: bool = True
    ) -> TemplateAnalysis:
        """Statically analyze this template and any included/rendered templates.

        Currently we only analyze references to template variables.

        :param follow_partials: If ``True``, we will try to load partial templates and
            analyze those templates too. Default's to ``True``.
        :type follow_partials: bool
        :param raise_for_failures: If ``True``, will raise an exception if an
            ``ast.Node`` or ``expression.Expression`` does not define a ``children()``
            method, or if a partial template can not be loaded. When ``False``, no
            exception is raised and a mapping of failed nodes and expressions is
            available as the ``failed_visits`` property. A mapping of unloadable partial
            templates is stored in the ``unloadable_partials`` property.
        :type raised_for_failed_visits: bool
        :returns: A object containing analysis results.
        :rtype: :class:`liquid.template.TemplateAnalysis`
        """
        refs = _TemplateVariableCounter(
            self,
            follow_partials=follow_partials,
            raise_for_failures=raise_for_failures,
        ).analyze()

        return TemplateAnalysis(
            variables=dict(refs.variables),
            local_variables=dict(refs.template_locals),
            global_variables=dict(refs.template_globals),
            failed_visits=dict(refs.failed_visits),
            unloadable_partials=dict(refs.unloadable_partials),
        )

    async def analyze_async(
        self, follow_partials: bool = True, raise_for_failures: bool = True
    ) -> TemplateAnalysis:
        """An async version of :meth:`analyze`."""
        refs = await _TemplateVariableCounter(
            self,
            follow_partials=follow_partials,
            raise_for_failures=raise_for_failures,
        ).analyze_async()

        return TemplateAnalysis(
            variables=dict(refs.variables),
            local_variables=dict(refs.template_locals),
            global_variables=dict(refs.template_globals),
            failed_visits=dict(refs.failed_visits),
            unloadable_partials=dict(refs.unloadable_partials),
        )

    def analyze_with_context(
        self, *args: Any, **kwargs: Any
    ) -> ContextualTemplateAnalysis:
        """Analyze a path through this template's syntax tree given some context data.

        Unlike :meth:`analyze`, this form of template analysis does perform a render,
        capturing template variables as we go.

        Python Liquid does not currently support template introspection from within a
        render context or ``Expression`` object. Meaning line numbers and template names
        are not available. Only variable names are reported along with the number of
        times they were referenced.

        It's also, using this method, not currently possible to detect names added to a
        block's scope. For example, ``forloop.index`` will be included in the results
        object if referenced within a ``for`` loop block.

        Accepts the same arguments as :meth:`render`.

        :returns: Contextual analysis results.
        :rtype: :class:`liquid.template.ContextualTemplateAnalysis`
        """
        context = self.capture_context_class(
            self.env, globals=self.make_globals(dict(*args, **kwargs))
        )
        buf = self._get_buffer()
        self.render_with_context(context, buf)
        return ContextualTemplateAnalysis(
            all_variables=dict(Counter(context.all_references)),
            local_variables=dict(Counter(context.local_references)),
            undefined_variables=dict(Counter(context.undefined_references)),
        )

    async def analyze_with_context_async(
        self, *args: Any, **kwargs: Any
    ) -> ContextualTemplateAnalysis:
        """An async version of :meth:`analyze_with_context`."""
        context = self.capture_context_class(
            self.env, globals=self.make_globals(dict(*args, **kwargs))
        )
        buf = self._get_buffer()
        await self.render_with_context_async(context, buf)
        return ContextualTemplateAnalysis(
            all_variables=dict(Counter(context.all_references)),
            local_variables=dict(Counter(context.local_references)),
            undefined_variables=dict(Counter(context.undefined_references)),
        )


class AwareBoundTemplate(BoundTemplate):
    """A `BoundTemplate` subclass that automatically includes a `TemplateDrop` in the
    global namespace."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.drop = TemplateDrop(self.name, self.path)

    def make_partial_namespace(
        self,
        partial: bool,
        render_args: Mapping[str, object],
    ) -> abc.Mapping[str, object]:
        return {
            "template": self.drop,
            **super().make_partial_namespace(partial, render_args),
        }


class TemplateDrop(Mapping[str, Optional[str]]):
    """Template meta data mapping."""

    def __init__(self, name: str, path: Optional[Union[str, Path]]):
        self.name = name
        self.path = path

        if not self.path or isinstance(self.path, str):
            self.path = Path(self.name)

        self.stem = self.path.stem
        self.suffix: Optional[str] = None

        if "." in self.stem:
            self.suffix = self.stem.split(".")[-1]

        self._items = {
            "directory": self.path.parent.name,
            "name": self.path.name.split(".")[0],
            "suffix": self.suffix,
        }

    def __str__(self) -> str:
        return self.stem

    def __repr__(self) -> str:
        return (
            f"TemplateDrop(directory='{self['directory']}', "
            f"name='{self['name']}', suffix='{self['suffix']}')"
        )  # pragma: no cover

    def __contains__(self, item: object) -> bool:
        return item in self._items

    def __getitem__(self, key: object) -> Optional[str]:
        return self._items[str(key)]

    def __len__(self) -> int:
        return len(self._items)

    def __iter__(self) -> Iterator[str]:
        return iter(self._items)


RE_SPLIT_IDENT = re.compile(r"(\.|\[)")

Refs = Dict[str, List[Tuple[str, int]]]
ReferenceMap = DefaultDict[str, List[Tuple[str, int]]]


# pylint: disable=too-few-public-methods
class ContextualTemplateAnalysis:
    """The result of analyzing a template's variables using
    :meth:`BoundTemplate.analyze_with_context`.

    Each of the following properties is a dictionary mapping variable names to the
    number of times that variable was referenced.

    :ivar all_variables: All variables references along a path through the template's
        syntax tree.
    :ivar local_variables: The names of variables assigned using the built-in ``assign``
        ``capture``, ``increment`` or ``decrement`` tags, or any custom tag that uses
        ``Context.assign()``.
    :ivar undefined_variables: The names of variables that could not be resolved. If a
        name is referenced before it is assigned, it will appear in ``undefined`` and
        ``assigns``.
    """

    __slots__ = ("all_variables", "local_variables", "undefined_variables")

    def __init__(
        self,
        *,
        all_variables: Dict[str, int],
        local_variables: Dict[str, int],
        undefined_variables: Dict[str, int],
    ) -> None:
        self.all_variables = all_variables
        self.local_variables = local_variables
        self.undefined_variables = undefined_variables


# pylint: disable=too-few-public-methods
class TemplateAnalysis:
    """The result of analyzing a template's variables using
    :meth:`BoundTemplate.analyze`.

    Each of the following properties is a dictionary mapping variable names to a list of
    two-tuples. Each tuple holds the location of a reference to the name as (<template
    name>, <line number>). If a name is referenced multiple times, it will appear
    multiple times in the list. If a name is referenced before it is "assigned", it will
    appear in ``local_variables`` and ``global_variables``.

    :ivar variables: All referenced variables, whether they are in scope or not.
        Including references to names such as ``forloop`` from the ``for`` tag.
    :ivar local_variables: Template variables that are added to the template local
        scope, whether they are subsequently used or not.
    :ivar global_variables: Template variables that, on the given line number and
        "file", are out of scope or are assumed to be "global". That is, expected to be
        included by the application developer rather than a template author.
    :ivar failed_visits: Names of AST ``Node`` and ``Expression`` objects that could not
        be visited, probably because they do not implement a ``children`` method.
    :ivar unloadable_partials: Names or identifiers of partial templates that could not
        be loaded. This will be empty if ``follow_partials`` is ``False``.
    """

    __slots__ = (
        "variables",
        "local_variables",
        "global_variables",
        "failed_visits",
        "unloadable_partials",
    )

    def __init__(
        self,
        *,
        variables: Refs,
        local_variables: Refs,
        global_variables: Refs,
        failed_visits: Refs,
        unloadable_partials: Refs,
    ) -> None:
        self.variables = variables
        self.local_variables = local_variables
        self.global_variables = global_variables
        self.failed_visits = failed_visits
        self.unloadable_partials = unloadable_partials


# pylint: disable=too-many-instance-attributes
class _TemplateVariableCounter:
    """Count references to variable names in a Liquid template.

    :param template: The Liquid template to analyze.
    :type template: liquid.BoundTemplate
    :param follow_partials: If ``True``, the reference counter will try to load partial
        templates and count variable references in those partials too. Default's to
        ``True``.
    :type follow_partials: bool
    :param raise_for_failures: If ``True``, will raise an exception if an ``ast.Node``
        or ``expression.Expression`` does not define a ``children()`` method, or if a
        partial template can not be loaded.

        When ``False``, no exception is raised and a mapping of failed nodes/expressions
        is available as the ``failed_visits`` property. A mapping of unloadable partial
        templates are stored in the ``unloadable_partials`` property.
    :type: raised_for_failed_visits: bool
    """

    def __init__(
        self,
        template: BoundTemplate,
        *,
        follow_partials: bool = True,
        raise_for_failures: bool = True,
        scope: Optional[ReadOnlyChainMap] = None,
        template_locals: Optional[ReferenceMap] = None,
        partials: Optional[List[Tuple[str, Optional[Dict[str, str]]]]] = None,
    ) -> None:
        self.template = template
        self._template_name = self.template.name or "<string>"

        self.follow_partials = follow_partials
        self.raise_for_failures = raise_for_failures

        # Names that are added to the template "local" scope.
        self.template_locals: ReferenceMap = (
            template_locals if template_locals is not None else defaultdict(list)
        )
        # Names that are referenced but are not in the template local scope
        self.template_globals: ReferenceMap = defaultdict(list)
        # Names that are referenced by a Liquid expression.
        self.variables: ReferenceMap = defaultdict(list)

        # Nodes and Expressions that don't implement a `children()` method.
        self.failed_visits: ReferenceMap = defaultdict(list)

        # Tags that load templates with an expression that can not be analyzed
        # statically.
        self.unloadable_partials: ReferenceMap = defaultdict(list)

        # Block scoped names.
        self._scope = scope if scope is not None else ReadOnlyChainMap()

        # Partial templates (include, render, etc.)
        self._partials = partials if partials is not None else []

        # get_template_with_context requires a `Context`.
        self._empty_context = Context(self.template.env)

    def analyze(self) -> _TemplateVariableCounter:
        """Traverse the template's syntax tree and count variables as we go.

        It is not safe to call this method multiple times.
        """
        for node in self.template.tree.statements:
            self._analyze(node)

        self._raise_for_failures()
        return self

    async def analyze_async(self) -> _TemplateVariableCounter:
        """An async version of :meth:`_TemplateVariableCounter.analyze`"""
        for node in self.template.tree.statements:
            await self._analyze_async(node)

        self._raise_for_failures()
        return self

    def _analyze(self, root: Node) -> None:
        try:
            children = root.children()
        except NotImplementedError:
            name = root.__class__.__name__
            self.failed_visits[name].append((self._template_name, root.token().linenum))
            return

        for child in children:
            self._analyze_child(child)

            if child.block_scope:
                self._scope.push({n: None for n in child.block_scope})

            if self.follow_partials:
                if child.load_mode == "include":
                    self._analyze_include(child)
                elif child.load_mode == "render":
                    self._analyze_render(child)
                elif child.load_mode is not None:
                    raise TemplateTraversalError(
                        f"unknown load mode '{child.load_mode}'"
                    )

            # Recurse
            if child.node:
                self._analyze(child.node)

            if child.block_scope:
                self._scope.pop()

    async def _analyze_async(self, root: Node) -> None:
        try:
            children = root.children()
        except NotImplementedError:
            name = root.__class__.__name__
            self.failed_visits[name].append((self._template_name, root.token().linenum))
            return

        for child in children:
            self._analyze_child(child)

            if child.block_scope:
                self._scope.push({n: None for n in child.block_scope})

            if self.follow_partials:
                if child.load_mode == "include":
                    await self._analyze_include_async(child)
                elif child.load_mode == "render":
                    await self._analyze_render_async(child)
                elif child.load_mode is not None:
                    raise TemplateTraversalError(
                        f"unknown load mode '{child.load_mode}'"
                    )

            # Recurse
            if child.node:
                await self._analyze_async(child.node)

            if child.block_scope:
                self._scope.pop()

    def _analyze_child(self, child: ChildNode) -> None:
        if child.expression:
            try:
                refs = self._analyze_expression(child.expression)
            except NotImplementedError:
                name = child.expression.__class__.__name__
                self.failed_visits[name].append((self._template_name, child.linenum))
                return

            for ref in refs:
                self.variables[ref].append((self._template_name, child.linenum))

            # Check refs that are not in scope or in the local namespace before
            # pushing the next block scope. This should highlight names that are
            # expected to be "global".
            for ref in refs:
                _ref = RE_SPLIT_IDENT.split(ref, 1)[0]
                if _ref not in self._scope and _ref not in self.template_locals:
                    self.template_globals[ref].append(
                        (self._template_name, child.linenum)
                    )

        if child.template_scope:
            for name in child.template_scope:
                self.template_locals[name].append((self._template_name, child.linenum))

    def _analyze_expression(self, expression: Expression) -> List[str]:
        """Return a list of references used in the given expression."""
        refs: List[str] = []

        if isinstance(expression, Identifier):
            refs.append(str(expression))

        for expr in expression.children():
            refs.extend(self._analyze_expression(expr))

        return refs

    def _analyze_include(self, child: ChildNode) -> None:
        name, load_context = self._include_context(child)
        if name is None or load_context is None:
            return

        try:
            template = self._empty_context.get_template_with_context(
                name, **load_context
            )
        except TemplateNotFound:
            self.unloadable_partials[name].append((self._template_name, child.linenum))
            return

        # Partial templates rendered in "include" mode share the same template local
        # namespace as their parent template. Note that block scoped variables have
        # already been pushed and will be popped by the caller.
        refs = _TemplateVariableCounter(
            template,
            follow_partials=self.follow_partials,
            scope=self._scope,
            template_locals=self.template_locals,
            partials=self._partials,
        ).analyze()

        self._update_reference_counters(refs)

    async def _analyze_include_async(self, child: ChildNode) -> None:
        name, load_context = self._include_context(child)
        if name is None or load_context is None:
            return

        try:
            template = await self._empty_context.get_template_with_context_async(
                name, **load_context
            )
        except TemplateNotFound:
            self.unloadable_partials[name].append((self._template_name, child.linenum))
            return

        refs = await _TemplateVariableCounter(
            template,
            follow_partials=self.follow_partials,
            scope=self._scope,
            template_locals=self.template_locals,
            partials=self._partials,
        ).analyze_async()

        self._update_reference_counters(refs)

    def _include_context(
        self, child: ChildNode
    ) -> Tuple[Optional[str], Optional[Dict[str, str]]]:
        # Partial templates rendered in "include" mode might use a variable template
        # name. We can't statically analyze a partial template unless it's name is a
        # literal string (or possibly an integer, but unlikely).
        if not isinstance(child.expression, Literal):
            key = str(child.expression)
            self.unloadable_partials[key].append((self._template_name, child.linenum))
            return None, None

        name = str(child.expression.value)
        load_context = child.load_context or {}

        # Keep track of partial templates that have already been analyzed. This prevents
        # us from analysing the same template twice and protects us against recursive
        # includes/renders.
        if (name, load_context) in self._partials:
            return None, None

        self._partials.append((name, load_context))
        return (name, load_context)

    def _analyze_render(self, child: ChildNode) -> None:
        name, load_context = self._render_context(child)
        if name is None or load_context is None:
            return

        try:
            template = self._empty_context.get_template_with_context(
                name, **load_context
            )
        except TemplateNotFound:
            self.unloadable_partials[name].append((self._template_name, child.linenum))
            return

        # Partial templates rendered in "render" mode do not share the parent template
        # local namespace. We do not pass the current block scope stack to "rendered"
        # templates either.
        scope = {n: None for n in child.block_scope} if child.block_scope else {}
        refs = _TemplateVariableCounter(
            template,
            follow_partials=self.follow_partials,
            scope=ReadOnlyChainMap(scope),
            partials=self._partials,
        ).analyze()

        self._update_reference_counters(refs)

    async def _analyze_render_async(self, child: ChildNode) -> None:
        name, load_context = self._render_context(child)
        if name is None or load_context is None:
            return None

        try:
            template = await self._empty_context.get_template_with_context_async(
                name, **load_context
            )
        except TemplateNotFound:
            self.unloadable_partials[name].append((self._template_name, child.linenum))
            return

        # Partial templates rendered in "render" mode do not share the parent template
        # local namespace. We do not pass the current block scope stack to "rendered"
        # templates either.
        scope = {n: None for n in child.block_scope} if child.block_scope else {}
        refs = await _TemplateVariableCounter(
            template,
            follow_partials=self.follow_partials,
            scope=ReadOnlyChainMap(scope),
            partials=self._partials,
        ).analyze_async()

        self._update_reference_counters(refs)

    def _render_context(
        self, child: ChildNode
    ) -> Tuple[Optional[str], Optional[Dict[str, str]]]:
        if not isinstance(child.expression, StringLiteral):
            raise TemplateTraversalError(
                "can't load from a variable expression when in 'render' mode"
            )

        name = child.expression.value
        load_context = child.load_context or {}

        # Keep track of partial templates that have already been analyzed. This prevents
        # us from analysing the same template twice and protects us against recursive
        # includes/renders.
        if (name, load_context) in self._partials:
            return None, None

        self._partials.append((name, load_context))
        return name, load_context

    def _update_reference_counters(self, refs: _TemplateVariableCounter) -> None:
        # Accumulate references from the partial/child template into its parent.
        for _name, _refs in refs.variables.items():
            self.variables[_name].extend(_refs)

        for _name, _refs in refs.template_globals.items():
            self.template_globals[_name].extend(_refs)

        for _name, _refs in refs.failed_visits.items():
            self.failed_visits[_name].extend(_refs)

        for _name, _refs in refs.unloadable_partials.items():
            self.unloadable_partials[_name].extend(_refs)

    def _raise_for_failures(self) -> None:
        if self.raise_for_failures and self.failed_visits:
            msg_target = next(iter(self.failed_visits.keys()))
            if len(self.failed_visits) > 1:
                msg = (
                    f"{msg_target} (+{len(self.failed_visits) -1} more) "
                    "does not implement a 'children' method"
                )
            else:
                msg = f"{msg_target} does not implement a 'children' method"
            raise TemplateTraversalError(f"failed visit: {msg}")

        if self.raise_for_failures and self.unloadable_partials:
            msg_target = next(iter(self.unloadable_partials.keys()))
            if len(self.unloadable_partials) > 1:
                msg = (
                    f"partial template '{msg_target}' "
                    f"(+{len(self.unloadable_partials) -1} more) "
                    "could not be loaded"
                )
            else:
                msg = f"partial template '{msg_target}' could not be loaded"
            raise TemplateTraversalError(f"failed visit: {msg}")



"""Tag and node definition for the built-in "liquid" tag."""

import sys

from typing import List
from typing import Optional
from typing import TextIO
from typing import TYPE_CHECKING

from liquid.ast import ChildNode
from liquid.ast import Node
from liquid.ast import BlockNode

from liquid.context import Context
from liquid.lex import get_liquid_expression_lexer

from liquid.parse import expect
from liquid.parse import get_parser

from liquid.stream import TokenStream
from liquid.tag import Tag

from liquid.token import Token
from liquid.token import TOKEN_TAG
from liquid.token import TOKEN_EXPRESSION
from liquid.token import TOKEN_EOF


if TYPE_CHECKING:  # pragma: no cover
    from liquid import Environment

TAG_LIQUID = sys.intern("liquid")


class LiquidNode(Node):
    """Parse tree node for the built-in "liquid" tag."""

    __slots__ = ("tok", "block")

    def __init__(self, tok: Token, block: BlockNode):
        self.tok = tok
        self.block = block

    def __str__(self) -> str:
        return str(self.block)

    def __repr__(self) -> str:  # pragma: no cover
        return f"LiquidNode(tok={self.tok}, block={self.block!r})"

    def render_to_output(self, context: Context, buffer: TextIO) -> Optional[bool]:
        return self.block.render(context, buffer)

    async def render_to_output_async(
        self, context: Context, buffer: TextIO
    ) -> Optional[bool]:
        return await self.block.render_async(context, buffer)

    def children(self) -> List[ChildNode]:
        return self.block.children()


class LiquidTag(Tag):
    """The built-in "liquid" tag."""

    name = TAG_LIQUID
    block = False

    def __init__(self, env: "Environment"):
        super().__init__(env)
        self.tokenize_liquid_expression = get_liquid_expression_lexer(
            comment_start_string=env.comment_start_string
        )

    def parse(self, stream: TokenStream) -> Node:
        expect(stream, TOKEN_TAG, value=TAG_LIQUID)
        tok = stream.current
        stream.next_token()

        if stream.current.type == TOKEN_EOF:
            # Empty liquid tag. Empty block.
            block = BlockNode(tok, [])
        else:
            expect(stream, TOKEN_EXPRESSION)
            expr_stream = TokenStream(
                self.tokenize_liquid_expression(
                    stream.current.value, line_count=stream.current.linenum
                )
            )
            parser = get_parser(self.env)
            block = parser.parse_block(expr_stream, end=())

        return LiquidNode(tok, block=block)



"""Golden tests cases for testing liquid's `for` tag."""

from liquid.golden.case import Case

cases = [
    Case(
        description="simple range loop",
        template=r"{% for i in (0..3) %}{{ i }} {% endfor %}",
        expect="0 1 2 3 ",
    ),
    Case(
        description="range loop using identifier",
        template=(
            r"{% for i in (0..product.end_range) %}"
            r"{{ i }} - {{ product.tags[i] }} "
            r"{% endfor %}"
        ),
        expect="0 - sports 1 - garden ",
        globals={"product": {"tags": ["sports", "garden"], "end_range": 1}},
    ),
    Case(
        description="simple array loop",
        template=r"{% for tag in product.tags %}{{ tag }} {% endfor %}",
        expect="sports garden ",
        globals={"product": {"tags": ["sports", "garden"]}},
    ),
    Case(
        description="loop over an array in reverse",
        template=r"{% for tag in product.tags reversed %}{{ tag }} {% endfor %}",
        expect="garden sports ",
        globals={"product": {"tags": ["sports", "garden"]}},
    ),
    Case(
        description="simple hash loop",
        template=r"{% for c in collection %}{{ c[0] }} {{ c[1] }} {% endfor %}",
        expect="title foo description bar ",
        globals={"collection": {"title": "foo", "description": "bar"}},
    ),
    Case(
        description="empty array with default",
        template=(
            r"{% for img in emptythings.array %}"
            r"{{ img.url }} "
            r"{% else %}"
            r"no images"
            r"{% endfor %}"
        ),
        expect="no images",
        globals={"emptythings": {"array": [], "map": {}, "string": ""}},
    ),
    Case(
        description="break",
        template=(
            r"{% for tag in product.tags %}"
            r"{% if tag == 'sports' %}"
            r"{% break %}"
            r"{% else %}"
            r"{{ tag }} "
            r"{% endif %}"
            r"{% else %}"
            r"no images"
            r"{% endfor %}"
        ),
        expect="",
        globals={"product": {"tags": ["sports", "garden"]}},
    ),
    Case(
        description="continue",
        template=(
            r"{% for tag in product.tags %}"
            r"{% if tag == 'sports' %}"
            r"{% continue %}"
            r"{% else %}"
            r"{{ tag }} "
            r"{% endif %}"
            r"{% else %}"
            r"no images"
            r"{% endfor %}"
        ),
        expect="garden ",
        globals={"product": {"tags": ["sports", "garden"]}},
    ),
    Case(
        description="limit",
        template=r"{% for tag in product.tags limit:1 %}{{ tag }} {% endfor %}",
        expect="sports ",
        globals={"product": {"tags": ["sports", "garden"]}},
    ),
    Case(
        description="offset",
        template=r"{% for tag in product.tags offset:1 %}{{ tag }} {% endfor %}",
        expect="garden ",
        globals={"product": {"tags": ["sports", "garden"]}},
    ),
    Case(
        description="forloop length",
        template=r"{% for tag in product.tags %}{{ forloop.length }} {% endfor %}",
        expect="2 2 ",
        globals={"product": {"tags": ["sports", "garden"]}},
    ),
    Case(
        description="forloop length with limit",
        template=r"{% for tag in tags limit:3 %}{{ forloop.length }} {% endfor %}",
        expect="3 3 3 ",
        globals={
            "tags": [
                "sports",
                "garden",
                "home",
                "diy",
                "motoring",
                "fashion",
            ]
        },
    ),
    Case(
        description="forloop length with offset",
        template=r"{% for tag in tags offset:3 %}{{ forloop.length }} {% endfor %}",
        expect="3 3 3 ",
        globals={
            "tags": [
                "sports",
                "garden",
                "home",
                "diy",
                "motoring",
                "fashion",
            ]
        },
    ),
    Case(
        description="forloop goes out of scope",
        template=(
            r"{% for tag in product.tags %}"
            r"{{ forloop.length }} "
            r"{% endfor %}"
            r"{{ forloop.length }}"
        ),
        expect="2 2 ",
        globals={"product": {"tags": ["sports", "garden"]}},
    ),
    Case(
        description="forloop.first",
        template=r"{% for tag in product.tags %}{{ forloop.first }} {% endfor %}",
        expect="true false ",
        globals={"product": {"tags": ["sports", "garden"]}},
    ),
    Case(
        description="forloop.last",
        template=r"{% for tag in product.tags %}{{ forloop.last }} {% endfor %}",
        expect="false true ",
        globals={"product": {"tags": ["sports", "garden"]}},
    ),
    Case(
        description="forloop.index",
        template=r"{% for tag in product.tags %}{{ forloop.index }} {% endfor %}",
        expect="1 2 ",
        globals={"product": {"tags": ["sports", "garden"]}},
    ),
    Case(
        description="forloop.index0",
        template=r"{% for tag in product.tags %}{{ forloop.index0 }} {% endfor %}",
        expect="0 1 ",
        globals={"product": {"tags": ["sports", "garden"]}},
    ),
    Case(
        description="forloop.rindex",
        template=r"{% for tag in product.tags %}{{ forloop.rindex }} {% endfor %}",
        expect="2 1 ",
        globals={"product": {"tags": ["sports", "garden"]}},
    ),
    Case(
        description="forloop.rindex0",
        template=r"{% for tag in product.tags %}{{ forloop.rindex0 }} {% endfor %}",
        expect="1 0 ",
        globals={"product": {"tags": ["sports", "garden"]}},
    ),
    Case(
        description="forloop name",
        template=r"{% for tag in product.tags limit:1 %}{{ forloop.name }}{% endfor %}",
        expect="tag-product.tags",
        globals={"product": {"tags": ["sports", "garden"]}},
    ),
    Case(
        description="forloop name of a range",
        template=r"{% for i in (1..3) limit:1 %}{{ forloop.name }}{% endfor %}",
        expect="i-(1..3)",
        globals={},
    ),
    Case(
        description="forloop no such attribute",
        template=r"{% for tag in product.tags %}{{ forloop.nosuchthing }}{% endfor %}",
        expect="",
        globals={"product": {"tags": ["sports", "garden"]}},
    ),
    Case(
        description="iterate an empty array",
        template=r"{% for item in emptythings.array %}{{ item }}{% endfor %}",
        expect="",
        globals={"emptythings": {"array": [], "map": {}, "string": ""}},
    ),
    Case(
        description="iterate an empty array with default",
        template=(
            r"{% for item in emptythings.array %}"
            r"{{ item }}{% else %}foo"
            r"{% endfor %}"
        ),
        expect="foo",
        globals={"emptythings": {"array": [], "map": {}, "string": ""}},
    ),
    Case(
        description="lookup a filter from an outer context",
        template=r"{% for tag in product.tags %}{{ tag | upcase }} {% endfor %}",
        expect="SPORTS GARDEN ",
        globals={"product": {"tags": ["sports", "garden"]}},
    ),
    Case(
        description="assign inside loop",
        template=(
            r"{% for tag in product.tags %}"
            r"{% assign x = tag %}"
            r"{% endfor %}"
            r"{{ x }}"
        ),
        expect="garden",
        globals={"product": {"tags": ["sports", "garden"]}},
    ),
    Case(
        description="blank empty loops",
        template=r"{% for i in (0..10) %}  {% endfor %}",
        expect="",
    ),
    Case(
        description=(
            "loop over nested and chained object from context "
            "with trailing identifier"
        ),
        template=(
            r"{% for link in linklists[section.settings.menu].links %}"
            r"{{ link }} "
            r"{% endfor %}"
        ),
        expect="1 2 ",
        globals={
            "linklists": {"main": {"links": ["1", "2"]}},
            "section": {"settings": {"menu": "main"}},
        },
    ),
    Case(
        description="loop over undefined",
        template=r"{% for tag in nosuchthing %}{{ tag }}{% endfor %}",
        expect="",
        globals={},
    ),
    Case(
        description="continue a loop",
        template=(
            r"{% for item in array limit: 3 %}"
            r"a{{ item }} "
            r"{% endfor %}"
            r"{% for item in array offset: continue %}"
            r"b{{ item }} "
            r"{% endfor %}"
        ),
        expect="a1 a2 a3 b4 b5 b6 ",
        globals={"array": [1, 2, 3, 4, 5, 6]},
    ),
    Case(
        description="continue a loop over an assigned range",
        template=(
            r"{% assign nums = (1..5) %}"
            r"{% for item in nums limit: 3 %}"
            r"a{{ item }} "
            r"{% endfor %}"
            r"{% for item in nums offset: continue %}"
            r"b{{ item }} "
            r"{% endfor %}"
        ),
        expect="a1 a2 a3 b4 b5 ",
    ),
    Case(
        description="continue a loop over a changing array",
        template=(
            r"{% assign foo = '1,2,3,4,5,6' | split: ',' %}"
            r"{% for item in foo limit: 3 %}"
            r"{{ item }} "
            r"{% endfor %}"
            r"{% assign foo = 'u,v,w,x,y,z' | split: ',' %}"
            r"{% for item in foo offset: continue %}"
            r"{{ item }} "
            r"{% endfor %}"
        ),
        expect="1 2 3 x y z ",
        globals={},
    ),
    Case(
        description="continue with changing loop var",
        template=(
            r"{% for foo in array limit: 3 %}"
            r"{{ foo }} "
            r"{% endfor %}"
            r"{% for bar in array offset: continue %}"
            r"{{ bar }} "
            r"{% endfor %}"
        ),
        expect="1 2 3 1 2 3 4 5 6 ",
        globals={"array": [1, 2, 3, 4, 5, 6]},
    ),
    Case(
        description="nothing to continue from",
        template=(
            r"{% for item in array %}"
            r"a{{ item }} "
            r"{% endfor %}"
            r"{% for item in array offset: continue %}"
            r"b{{ item }} "
            r"{% endfor %}"
        ),
        expect="a1 a2 a3 a4 a5 a6 ",
        globals={"array": [1, 2, 3, 4, 5, 6]},
    ),
    Case(
        description="offset continue without preceding loop",
        template=r"{% for item in array offset: continue %}{{ item }} {% endfor %}",
        expect="1 2 3 4 5 6 ",
        globals={"array": [1, 2, 3, 4, 5, 6]},
    ),
    Case(
        description="continue from a limit that is greater than length",
        template=(
            r"{% for item in array limit: 99 %}"
            r"a{{ item }} "
            r"{% endfor %}"
            r"{% for item in array offset: continue %}"
            r"b{{ item }} "
            r"{% endfor %}"
        ),
        expect="a1 a2 a3 a4 a5 a6 ",
        globals={"array": [1, 2, 3, 4, 5, 6]},
    ),
    Case(
        description="continue from a range expression",
        template=(
            r"{% for item in (1..6) limit: 3 %}"
            r"a{{ item }} "
            r"{% endfor %}"
            r"{% for item in (1..6) offset: continue %}"
            r"b{{ item }} "
            r"{% endfor %}"
        ),
        expect="a1 a2 a3 b4 b5 b6 ",
        globals={"array": [1, 2, 3, 4, 5, 6]},
    ),
    Case(
        description="offset continue twice with limit",
        template=(
            r"{% for item in (1..6) limit: 2 %}"
            r"a{{ item }} "
            r"{% endfor %}"
            r"{% for item in (1..6) limit: 2 offset: continue %}"
            r"b{{ item }} "
            r"{% endfor %}"
            r"{% for item in (1..6) offset: continue %}"
            r"c{{ item }} "
            r"{% endfor %}"
        ),
        expect="a1 a2 b3 b4 c5 c6 ",
    ),
    Case(
        description="offset continue twice with changing limit",
        template=(
            r"{% for item in (1..6) limit: 2 %}"
            r"a{{ item }} "
            r"{% endfor %}"
            r"{% for item in (1..6) limit: 3 offset: continue %}"
            r"b{{ item }} "
            r"{% endfor %}"
            r"{% for item in (1..6) offset: continue %}"
            r"c{{ item }} "
            r"{% endfor %}"
        ),
        expect="a1 a2 b3 b4 b5 c6 ",
    ),
    Case(
        description="offset continue twice with no second limit",
        template=(
            r"{% for item in (1..6) limit: 2 %}"
            r"a{{ item }} "
            r"{% endfor %}"
            r"{% for item in (1..6) offset: continue %}"
            r"b{{ item }} "
            r"{% endfor %}"
            r"{% for item in (1..6) offset: continue %}"
            r"c{{ item }} "
            r"{% endfor %}"
        ),
        expect="a1 a2 b3 b4 b5 b6 ",
    ),
    Case(
        description="offset continue from a broken loop",
        template=(
            r"{% for item in (1..6) limit: 4 %}"
            r"{% if item == 3 %}{% break %}{% endif %}"
            r"a{{ item }} "
            r"{% endfor %}"
            r"{% for item in (1..6) offset: continue %}"
            r"b{{ item }} "
            r"{% endfor %}"
        ),
        expect="a1 a2 b5 b6 ",
    ),
    Case(
        description="offset continue from a broken loop with preceding limit",
        template=(
            r"{% for item in (1..6) limit: 3 %}"
            r"a{{ item }} "
            r"{% endfor %}"
            r"{% for item in (1..6) %}"
            r"{% if item == 3 %}{% break %}{% endif %}"
            r"b{{ item }} "
            r"{% endfor %}"
            r"{% for item in (1..6) offset: continue %}"
            r"c{{ item }} "
            r"{% endfor %}"
        ),
        expect="a1 a2 a3 b1 b2 ",
    ),
    Case(
        description="offset continue forloop length",
        template=(
            r"{% for item in (1..6) limit: 2 %}"
            r"a{{ item }} - {{ forloop.length }}, "
            r"{% endfor %}"
            r"{% for item in (1..6) offset: continue %}"
            r"b{{ item }} - {{ forloop.length }}, "
            r"{% endfor %}"
        ),
        expect="a1 - 2, a2 - 2, b3 - 4, b4 - 4, b5 - 4, b6 - 4, ",
    ),
    Case(
        description="parentloop is normally undefined",
        template=(r"{% for i in (1..2)%}{{ forloop.parentloop.index }}{% endfor %}"),
        expect="",
        globals={},
    ),
    Case(
        description="access parentloop",
        template=(
            r"{% for i in (1..2)%}"
            r"{% for j in (1..2) %}"
            r"{{ i }} {{j}} {{ forloop.parentloop.index }} {{ forloop.index }} "
            r"{% endfor %}"
            r"{% endfor %}"
        ),
        expect="1 1 1 1 1 2 1 2 2 1 2 1 2 2 2 2 ",
        globals={},
    ),
    Case(
        description="parentloop goes out of scope",
        template=(
            r"{% for i in (1..2)%}"
            r"{% for j in (1..2) %}"
            r"{{ i }} {{ j }} "
            r"{% endfor %}"
            r"{{ forloop.parentloop.index }}"
            r"{% endfor %}"
        ),
        expect="1 1 1 2 2 1 2 2 ",
        globals={},
    ),
    Case(
        description="parent's parentloop",
        template=(
            r"{% for i in (1..2) %}"
            r"{% for j in (1..2) %}"
            r"{% for k in (1..2) %}"
            r"i={{ forloop.parentloop.parentloop.index }} "
            r"j={{ forloop.parentloop.index }} "
            r"k={{ forloop.index }} "
            r"{% endfor %}"
            r"{% endfor %}"
            r"{% endfor %}"
        ),
        expect=(
            "i=1 j=1 k=1 i=1 j=1 k=2 "
            "i=1 j=2 k=1 i=1 j=2 k=2 "
            "i=2 j=1 k=1 i=2 j=1 k=2 "
            "i=2 j=2 k=1 i=2 j=2 k=2 "
        ),
        globals={},
    ),
    Case(
        description="loop over an existing range object",
        template=(
            r"{% assign foo = (1..3) %}"
            r"{{ foo | join: '#' }}"
            r"{% for i in foo %}"
            r"{{ i }}"
            r"{% endfor %}"
            r"{% for i in foo %}"
            r"{{ i }}"
            r"{% endfor %}"
        ),
        expect="1#2#3123123",
        globals={},
    ),
    Case(
        description="loop over range with float start",
        template=r"{% assign x = (2.4..5) %}{% for i in x %}{{ i }}{% endfor %}",
        expect="2345",
        globals={},
    ),
    Case(
        description="share outer scope",
        template=(
            r"{% assign foo = 'hello' %}"
            r"{% for x in (1..3) %}"
            r"{% assign foo = x %}"
            r"{% endfor %}"
            r"{{ foo }}"
        ),
        expect="3",
        globals={},
    ),
    Case(
        description="offset and limit",
        template=r"{% for tag in tags limit: 3 offset: 1 %}{{ tag }} {% endfor %}",
        expect="garden home diy ",
        globals={"tags": ["sports", "garden", "home", "diy", "motoring", "fashion"]},
    ),
    Case(
        description="first and last with an offset and limit",
        template=(
            r"{% for tag in tags limit: 2 offset: 1 %}"
            r"{{ tag }} {{ forloop.first }} {{ forloop.last }} "
            r"{% endfor %}"
        ),
        expect="garden true false home false true ",
        globals={"tags": ["sports", "garden", "home", "diy", "motoring", "fashion"]},
    ),
    Case(
        description="first and last with offset continue",
        template=(
            r"{% for tag in product.tags limit: 1 %}"
            r"{% endfor %}"
            r"{% for tag in product.tags offset: continue %}"
            r"{{ forloop.first }} {{ forloop.last }} "
            r"{% endfor %}"
        ),
        expect="true false false false false false false false false true ",
        globals={
            "product": {
                "tags": ["sports", "garden", "home", "diy", "motoring", "fashion"]
            }
        },
    ),
    Case(
        description="range start and stop are the same",
        template=r"{% for i in (1..1) %}{{ i }} {% endfor %}",
        expect="1 ",
    ),
    Case(
        description="range start and stop are zero",
        template=r"{% for i in (0..0) %}{{ i }} {% endfor %}",
        expect="0 ",
    ),
    # Case(
    #     description="limit is a float",
    #     template=r"{% for i in (1..4) limit: 2.6 %}{{ i }} {% endfor %}",
    #     expect="",
    #     error=True,
    # ),
    # Case(
    #     description="offset is a float",
    #     template=r"{% for i in (1..4) offset: 2.6 %}{{ i }} {% endfor %}",
    #     expect="",
    #     error=True,
    # ),
    Case(
        description="limit is a string",
        template=r"{% for i in (1..4) limit: '2' %}{{ i }} {% endfor %}",
        expect="1 2 ",
    ),
    Case(
        description="offset is a string",
        template=r"{% for i in (1..4) offset: '2' %}{{ i }} {% endfor %}",
        expect="3 4 ",
    ),
    Case(
        description="limit is a non-number string",
        template=r"{% for i in (1..4) limit: 'foo' %}{{ i }} {% endfor %}",
        expect="",
        error=True,
    ),
    Case(
        description="offset is a non-number string",
        template=r"{% for i in (1..4) offset: 'foo' %}{{ i }} {% endfor %}",
        expect="",
        error=True,
    ),
    Case(
        description="limit is not a string or number",
        template=r"{% for i in (1..4) limit: foo %}{{ i }} {% endfor %}",
        expect="",
        globals={"foo": [1, 2, 3]},
        error=True,
    ),
    Case(
        description="offset is not a string or number",
        template=r"{% for i in (1..4) offset: foo %}{{ i }} {% endfor %}",
        expect="",
        globals={"foo": [1, 2, 3]},
        error=True,
    ),
    Case(
        description="comma separated arguments",
        template=r"{% for i in (1..6), limit: 4, offset: 2 %}{{ i }} {% endfor %}",
        expect="3 4 5 6 ",
        globals={},
    ),
    Case(
        description="some comma separated arguments",
        template=r"{% for i in (1..6) limit: 4, offset: 2, %}{{ i }} {% endfor %}",
        expect="3 4 5 6 ",
        globals={},
    ),
]



"""Test cases for `macro` and `call` tags."""
import asyncio
from unittest import TestCase

from typing import Mapping
from typing import NamedTuple
from typing import Any
from typing import Dict

from liquid.context import StrictUndefined
from liquid.environment import Environment
from liquid.exceptions import UndefinedError
from liquid.extra import add_macro_tags
from liquid.loaders import DictLoader


class Case(NamedTuple):
    """Table-driven test helper."""

    description: str
    template: str
    expect: str
    globals: Mapping[str, Any]
    partials: Dict[str, str]


class MacroRenderTestCase(TestCase):
    """Test cases for rendering the `macro` and `call` tags."""

    def test_render_macro(self) -> None:
        """Test that we can render `macro` and `call` tags."""
        test_cases = [
            Case(
                description="basic macro no call",
                template=r"{% macro 'func' %}Hello, World!{% endmacro %}",
                expect="",
                globals={},
                partials={},
            ),
            Case(
                description="call basic macro",
                template=(
                    r"{% macro 'func' %}Hello, World!{% endmacro %}"
                    r"{% call 'func' %}"
                ),
                expect="Hello, World!",
                globals={},
                partials={},
            ),
            Case(
                description="call basic macro multiple times",
                template=(
                    r"{% macro 'func' %}Hello, World!{% endmacro %}"
                    r"{% call 'func' %}"
                    r"{% call 'func' %}"
                ),
                expect="Hello, World!Hello, World!",
                globals={},
                partials={},
            ),
            Case(
                description="call macro with argument",
                template=(
                    r"{% macro 'func', you %}Hello, {{ you }}!{% endmacro %}"
                    r"{% call 'func', 'you' %} "
                    r"{% call 'func', 'World' %}"
                ),
                expect="Hello, you! Hello, World!",
                globals={},
                partials={},
            ),
            Case(
                description="call macro with default argument",
                template=(
                    r"{% macro 'func' you: 'brian' %}Hello, {{ you }}!{% endmacro %}"
                    r"{% call 'func' %} "
                    r"{% call 'func' you: 'World' %}"
                ),
                expect="Hello, brian! Hello, World!",
                globals={},
                partials={},
            ),
            Case(
                description="boolean literal default argument",
                template=(
                    r"{% macro 'func' foo: false %}"
                    r"{% if foo %}Hello, World!{% endif %}"
                    r"{% endmacro %}"
                    r"{% call 'func' %}"
                    r"{% call 'func' foo: true %}"
                ),
                expect="Hello, World!",
                globals={},
                partials={},
            ),
            Case(
                description="chained default argument from context",
                template=(
                    r"{% macro 'func' greeting: foo.bar %}"
                    r"{{ greeting }}, World!"
                    r"{% endmacro %}"
                    r"{% call 'func' %}"
                    r"{% call 'func' greeting: 'Goodbye' %}"
                ),
                expect="Hello, World!Goodbye, World!",
                globals={"foo": {"bar": "Hello"}},
                partials={},
            ),
            Case(
                description="excess arguments",
                template=(
                    r"{% macro 'func' %}"
                    r"{{ args | join: '-' }}"
                    r"{% endmacro %}"
                    r"{% call 'func' 1, 2 %}"
                ),
                expect="1-2",
                globals={},
                partials={},
            ),
            Case(
                description="excess keyword arguments",
                template=(
                    r"{% macro 'func' %}"
                    r"{% for arg in kwargs %}"
                    r"{{ arg[0] }} => {{ arg[1] }}, "
                    r"{% endfor %}"
                    r"{% endmacro %}"
                    r"{% call 'func', a: 1, b: 2 %}"
                ),
                expect="a => 1, b => 2, ",
                globals={},
                partials={},
            ),
            Case(
                description="missing argument",
                template=(
                    r"{% macro 'func', foo %}"
                    r"{{ foo }}"
                    r"{% endmacro %}"
                    r"{% call 'func' %}"
                ),
                expect="",
                globals={},
                partials={},
            ),
            Case(
                description="undefined macro",
                template=(r"{% call 'func' %}"),
                expect="",
                globals={},
                partials={},
            ),
            Case(
                description="default before positional",
                template=(
                    r"{% macro 'func' you: 'brian', greeting %}"
                    r"{{ greeting }}, {{ you }}!"
                    r"{% endmacro %}"
                    r"{% call 'func' %} "
                    r"{% call 'func' you: 'World', greeting: 'Goodbye' %}"
                ),
                expect=", brian! Goodbye, World!",
                globals={},
                partials={},
            ),
        ]

        for case in test_cases:
            env = Environment(loader=DictLoader(case.partials))
            add_macro_tags(env)

            template = env.from_string(case.template, globals=case.globals)

            with self.subTest(msg=case.description):
                result = template.render()
                self.assertEqual(result, case.expect)

    def test_render_macro_strict_undefined(self) -> None:
        """Test that missing arguments and undefined macros are `Undefined`."""
        test_cases = [
            Case(
                description="missing argument",
                template=(
                    r"{% macro 'func', foo %}"
                    r"{{ foo }}"
                    r"{% endmacro %}"
                    r"{% call 'func' %}"
                ),
                expect="'foo' is undefined, on line 1",
                globals={},
                partials={},
            ),
            Case(
                description="undefined macro",
                template=(r"{% call 'func' %}"),
                expect="'func' is undefined, on line 1",
                globals={},
                partials={},
            ),
        ]

        for case in test_cases:
            env = Environment(
                loader=DictLoader(case.partials),
                undefined=StrictUndefined,
            )
            add_macro_tags(env)

            template = env.from_string(case.template, globals=case.globals)

            with self.subTest(msg=case.description):
                with self.assertRaises(UndefinedError) as raised:
                    template.render()
                self.assertEqual(case.expect, str(raised.exception))

    def test_render_macro_async(self) -> None:
        """Test that we can render a macro asynchronously."""
        env = Environment()
        add_macro_tags(env)

        template = env.from_string(
            r"{% macro 'foo', you: 'World' %}"
            r"Hello, {{ you }}!"
            r"{% endmacro %}"
            r"{% call 'foo' %}"
            r"{% call 'foo', you: 'you' %}"
            r"{% call 'nosuchthing' %}"
        )

        async def coro() -> str:
            return await template.render_async()

        result = asyncio.run(coro())
        self.assertEqual(result, "Hello, World!Hello, you!")


class AnalyzeMacroTestCase(TestCase):
    """Test cases for analyzing `macro` and `call` tags."""

    def test_analyze_macro_tag(self) -> None:
        """Test that we can statically analyze macro and call tags."""
        env = Environment()
        add_macro_tags(env)

        template = env.from_string(
            r"{% macro 'foo', you: 'World', arg: n %}"
            r"Hello, {{ you }}!"
            r"{% endmacro %}"
            r"{% call 'foo' %}"
            r"{% assign x = 'you' %}"
            r"{% call 'foo', you: x %}"
        )

        expected_template_globals = {
            "n": [("<string>", 1)],
        }
        expected_template_locals = {"x": [("<string>", 1)]}
        expected_refs = {
            "n": [("<string>", 1)],
            "you": [("<string>", 1)],
            "x": [("<string>", 1)],
        }

        analysis = template.analyze()

        self.assertEqual(analysis.local_variables, expected_template_locals)
        self.assertEqual(analysis.global_variables, expected_template_globals)
        self.assertEqual(analysis.variables, expected_refs)



"""Tag and node definition for the mock "paginate" tag."""
# pylint: disable=missing-function-docstring
import math
import sys

from collections import abc

from typing import List
from typing import Optional
from typing import TextIO

from liquid.token import Token
from liquid.token import TOKEN_EXPRESSION
from liquid.token import TOKEN_TAG
from liquid.token import TOKEN_BY
from liquid.token import TOKEN_IDENTIFIER
from liquid.token import TOKEN_INTEGER
from liquid.token import TOKEN_EOF

from liquid.tag import Tag
from liquid import ast
from liquid.stream import TokenStream
from liquid.lex import tokenize_paginate_expression
from liquid.expression import Identifier
from liquid.context import Context

from liquid.parse import get_parser
from liquid.parse import expect
from liquid.parse import parse_identifier
from liquid.parse import parse_integer_literal

from liquid.exceptions import LiquidTypeError


TAG_PAGINATE = sys.intern("paginate")
TAG_ENDPAGINATE = sys.intern("endpaginate")

END_PAGINATEBLOCK = frozenset((TAG_ENDPAGINATE,))


class PaginateNode(ast.Node):
    """Parse tree node for the mock "paginate" tag."""

    __slots__ = ("tok", "identifier", "page_size", "block")

    def __init__(
        self, tok: Token, identifier: Identifier, page_size: int, block: ast.BlockNode
    ):
        self.tok = tok
        self.identifier = identifier
        self.page_size = page_size
        self.block = block

    def __str__(self):
        return f"paginate({self.identifier} by {self.page_size}) {{ {self.block} }}"

    def render_to_output(self, context: Context, buffer: TextIO) -> Optional[bool]:
        collection = self.identifier.evaluate(context)
        assert isinstance(collection, abc.Collection)

        if not collection:
            raise LiquidTypeError(f"cannot paginate {self.identifier}")

        collection_size = len(collection)
        page_count = math.ceil(collection_size / self.page_size)

        current_page = context.get("current_page", default=1)
        assert isinstance(current_page, int)

        pagination = {
            "page_size": self.page_size,
            "current_page": current_page,
            "current_offset": current_page * self.page_size,
            "items": collection_size,
            "pages": page_count,
            "parts": [],
            "previous": None,
            "next": None,
        }

        if current_page > 1:
            pagination["previous"] = link("&laquo; Previous", current_page - 1)

        if current_page < page_count:
            pagination["next"] = link("Next &raquo;", current_page + 1)

        # NOTE: Not an accurate port of the reference implementation. But should
        # be close enough for a benchmark.
        if page_count > 1:
            for page in range(1, page_count + 1):
                if current_page == page:
                    pagination["parts"].append(no_link(page))
                else:
                    pagination["parts"].append(link(page, page))

        with context.extend({"paginate": pagination}):
            self.block.render(context, buffer)

    def children(self) -> List[ast.ChildNode]:
        return [
            ast.ChildNode(
                linenum=self.tok.linenum,
                expression=self.identifier,
                node=self.block,
                block_scope=["paginate"],
            )
        ]


def link(title, page):
    return {
        "title": title,
        "url": f"/collections/frontpage?page={page}",
        "is_link": True,
    }


def no_link(title):
    return {"title": title, "is_link": False}


class PaginateTag(Tag):
    """The mock "paginate" tag."""

    name = TAG_PAGINATE
    end = TAG_ENDPAGINATE

    def parse(self, stream: TokenStream) -> ast.Node:
        parser = get_parser(self.env)

        expect(stream, TOKEN_TAG, value=TAG_PAGINATE)
        tok = stream.current
        stream.next_token()

        expect(stream, TOKEN_EXPRESSION)
        expr_stream = TokenStream(tokenize_paginate_expression(stream.current.value))

        # Read the identifier (the object to be paginated).
        expect(expr_stream, TOKEN_IDENTIFIER)
        identifier = parse_identifier(expr_stream)
        expr_stream.next_token()

        # Eat TOKEN_BY
        expect(expr_stream, TOKEN_BY)
        expr_stream.next_token()

        # Read the number of items per page.
        expect(expr_stream, TOKEN_INTEGER)
        page_size = parse_integer_literal(expr_stream)
        expr_stream.next_token()

        # End of expression
        expect(expr_stream, TOKEN_EOF)

        # Advance the stream passed the expression and read the block.
        stream.next_token()
        block = parser.parse_block(stream, end=END_PAGINATEBLOCK)
        expect(stream, TOKEN_TAG, value=TAG_ENDPAGINATE)

        return PaginateNode(
            tok, identifier=identifier, page_size=page_size.value, block=block
        )



"""Golden tests cases for testing range objects."""

from liquid.golden.case import Case

cases = [
    Case(
        description="start is not a number",
        template=r"{{ (start..end) | join: '#' }}",
        globals={"start": "foo", "end": 5},
        expect="0#1#2#3#4#5",
    ),
    Case(
        description="end is not a number",
        template=r"{{ (start..end) | join: '#' }}",
        globals={"start": "1", "end": "foo"},
        expect="",
    ),
    Case(
        description="end is less than start",
        template=r"{{ (start..end) | join: '#' }}",
        globals={"start": 5, "end": 1},
        expect="",
    ),
    Case(
        description="start is negative",
        template=r"{{ (start..end) | join: '#' }}",
        globals={"start": -5, "end": 1},
        expect="-5#-4#-3#-2#-1#0#1",
    ),
    Case(
        description="start and end are negative",
        template=r"{{ (start..end) | join: '#' }}",
        globals={"start": -5, "end": -2},
        expect="-5#-4#-3#-2",
    ),
]



# pylint: disable=missing-module-docstring
from .array import index
from .html import script_tag
from .html import stylesheet_tag
from ._json import JSON


__all__ = (
    "index",
    "JSON",
    "script_tag",
    "stylesheet_tag",
)



"""Liquid specific Exceptions and warnings."""

from typing import Any
from typing import Dict
from typing import Optional
from typing import Type
from typing import Union

from pathlib import Path


class Error(Exception):
    """Base class for all Liquid exceptions."""

    def __init__(
        self,
        *args: object,
        linenum: Optional[int] = None,
        filename: Optional[Union[str, Path]] = None,
    ):
        self.linenum = linenum
        self.filename = filename
        super().__init__(*args)

    def __str__(self) -> str:
        msg = super().__str__()
        if self.linenum:
            msg = f"{msg}, on line {self.linenum}"
        if self.filename:
            msg += f" of {self.filename}"
        return msg

    @property
    def message(self) -> object:
        """Return the exception's error message if one was given."""
        if self.args:
            return self.args[0]
        return None


class LiquidInterrupt(Exception):
    """Loop interrupt"""


class LiquidSyntaxError(Error):
    """Exception raised when there is a parser error."""

    def __init__(
        self,
        *args: object,
        linenum: Optional[int] = None,
        filename: Optional[Union[str, Path]] = None,
    ):
        super().__init__(*args, linenum=linenum, filename=filename)
        self.source: Optional[str] = None

    @property
    def name(self) -> str:
        """Return the name of the template that raised this exception. Return an empty
        string if a name is not available."""
        if isinstance(self.filename, Path):
            return self.filename.as_posix()
        if self.filename:
            return str(self.filename)
        return ""


class LiquidTypeError(Error):
    """Exception raised when an error occurs at render time."""


class DisabledTagError(Error):
    """Exception raised when an attempt is made to render a disabled tag."""


class NoSuchFilterFunc(Error):
    """Exception raised when a filter lookup fails."""


class FilterError(Error):
    """Exception raised when a filter fails."""


class FilterArgumentError(Error):
    """Exception raised when a filter's arguments are invalid."""


class FilterValueError(Error):
    """Exception raised when a filters value is invalid."""


class TemplateNotFound(Error):
    """Exception raised when a template could not be found."""

    def __str__(self) -> str:
        msg = super().__str__()
        msg = f"template not found {msg}"
        return msg


class ResourceLimitError(Error):
    """Base class for exceptions relating to resource limits."""


class ContextDepthError(ResourceLimitError):
    """Exception raised when the maximum context depth is reached.

    Usually indicates recursive use of ``render`` or ``include`` tags.
    """


class LoopIterationLimitError(ResourceLimitError):
    """Exception raised when the maximum number of loop iterations has
    been exceeded.
    """


class OutputStreamLimitError(ResourceLimitError):
    """Exception raised when an output stream limit has been exceeded."""


class LocalNamespaceLimitError(ResourceLimitError):
    """Exception raised when the maximum size of a render context's local namespace
    has been exceeded."""


# LiquidValueError inheriting from LiquidSyntaxError does not make complete sense.
# The alternative is to have multiple to_int functions that raise more appropriate
# exceptions depending on whether we are parsing or rendering when attempting to
# convert long strings to integers.


class LiquidValueError(LiquidSyntaxError):
    """Exception raised when a cast from str to int exceeds the length limit."""


class UndefinedError(Error):
    """Exception raised by the StrictUndefined type."""


class BreakLoop(LiquidInterrupt):
    """Exception raised when a BreakNode is rendered."""


class ContinueLoop(LiquidInterrupt):
    """Exception raised when a ContinueNode is rendered."""


class TemplateTraversalError(Error):
    """Exception raised when an AST node or expression can not be visited."""


class LiquidWarning(UserWarning):
    """Base warning."""


class LiquidSyntaxWarning(LiquidWarning):
    """Replaces LiquidSyntaxError when in WARN mode."""


class LiquidTypeWarning(LiquidWarning):
    """Replaces LiquidTypeError when in WARN mode."""


class FilterWarning(LiquidWarning):
    """Replaces filter exceptions when in WARN mode."""


WARNINGS: Dict[Type[Error], Type[LiquidWarning]] = {
    LiquidSyntaxError: LiquidSyntaxWarning,
    LiquidTypeError: LiquidTypeWarning,
    FilterArgumentError: FilterWarning,
    NoSuchFilterFunc: FilterWarning,
}


def lookup_warning(exc: Type[Error]) -> Type[LiquidWarning]:
    """Return a warning equivalent of the given exception."""
    return WARNINGS.get(exc, LiquidWarning)


def escape(val: Any) -> str:
    """A dummy escape function that always raises an exception."""
    raise Error("autoescape requires Markupsafe to be installed")


class Markup(str):
    """A dummy markup class that always raises an exception."""

    def __init__(self, _: object):
        super().__init__()
        raise Error("autoescape requires Markupsafe to be installed")

    def join(self, _: object) -> str:
        raise Error(
            "autoescape requires Markupsafe to be installed"
        )  # pragma: no cover

    # pylint: disable=missing-function-docstring
    def unescape(self) -> str:
        raise Error(
            "autoescape requires Markupsafe to be installed"
        )  # pragma: no cover

    def format(self, *args: Any, **kwargs: Any) -> str:
        raise Error(
            "autoescape requires Markupsafe to be installed"
        )  # pragma: no cover



"""Golden tests cases for testing liquid's built-in `at_most` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="positive integer < arg",
        template=r"{{ 5 | at_most: 8 }}",
        expect="5",
    ),
    Case(
        description="positive integer > arg",
        template=r"{{ 8 | at_most: 5 }}",
        expect="5",
    ),
    Case(
        description="negative integer < arg",
        template=r"{{ -8 | at_most: 5 }}",
        expect="-8",
    ),
    Case(
        description="positive integer == arg",
        template=r"{{ 5 | at_most: 5 }}",
        expect="5",
    ),
    Case(
        description="positive float < arg",
        template=r"{{ 5.4 | at_most: 8.9 }}",
        expect="5.4",
    ),
    Case(
        description="positive float > arg",
        template=r"{{ 8.4 | at_most: 5.9 }}",
        expect="5.9",
    ),
    Case(
        description="positive string > arg",
        template=r'{{ "9" | at_most: 8 }}',
        expect="8",
    ),
    Case(
        description="missing arg",
        template=r"{{ 5 | at_most }}",
        expect="",
        error=True,
    ),
    Case(
        description="too many args",
        template=r"{{ 5 | at_most: 1, 2}}",
        expect="",
        error=True,
    ),
    Case(
        description="undefined left value",
        template=r"{{ nosuchthing | at_most: 5 }}",
        expect="0",
    ),
    Case(
        description="undefined argument",
        template=r"{{ 5 | at_most: nosuchthing }}",
        expect="0",
    ),
    Case(
        description="left value not a number",
        template=r'{{ "abc" | at_most: -2 }}',
        expect="-2",
    ),
]



"""Golden test cases for testing liquid identifiers."""

from liquid.golden.case import Case

cases = [
    Case(
        description="ascii lowercase",
        template="{% assign foo = 'hello' %}{{ foo }} {{ bar }}",
        expect="hello goodbye",
        globals={"bar": "goodbye"},
        strict=True,
    ),
    Case(
        description="ascii uppercase",
        template="{% assign FOO = 'hello' %}{{ FOO }} {{ BAR }}",
        expect="hello goodbye",
        globals={"BAR": "goodbye"},
        strict=True,
    ),
    Case(
        description="digits",
        template="{% assign foo1 = 'hello' %}{{ foo1 }} {{ bar2 }}",
        expect="hello goodbye",
        globals={"bar2": "goodbye"},
        strict=True,
    ),
    Case(
        description="only digits",
        template="{% assign 123 = 'hello' %}{{ 123 }} {{ 456 }}",
        expect="123 456",
        globals={"456": "goodbye"},
        strict=True,
    ),
    # Case(
    #     description="leading digits",
    #     template="{% assign 1foo = 'hello' %}{{ 1foo }} {{ 2bar }}",
    #     expect="hello goodbye",
    #     globals={"2bar": "goodbye"},
    #     strict=True,
    #     error=True,
    # ),
    Case(
        description="hyphens",
        template="{% assign foo-a = 'hello' %}{{ foo-a }} {{ bar-b }}",
        expect="hello goodbye",
        globals={"bar-b": "goodbye"},
        strict=True,
    ),
    Case(
        description="leading hyphen",
        template="{% assign -foo = 'hello' %}{{ -foo }} {{ -bar }}",
        expect="hello goodbye",
        globals={"-bar": "goodbye"},
        strict=True,
        error=True,
    ),
    Case(
        description="underscore",
        template="{% assign foo_a = 'hello' %}{{ foo_a }} {{ bar_b }}",
        expect="hello goodbye",
        globals={"bar_b": "goodbye"},
        strict=True,
    ),
    Case(
        description="only underscore",
        template="{% assign _ = 'hello' %}{{ _ }} {{ __ }}",
        expect="hello goodbye",
        globals={"__": "goodbye"},
        strict=True,
    ),
    Case(
        description="leading underscore",
        template="{% assign _foo = 'hello' %}{{ _foo }} {{ _bar }}",
        expect="hello goodbye",
        globals={"_bar": "goodbye"},
        strict=True,
    ),
    Case(
        description="trailing question mark assign",
        template="{% assign foo? = 'hello' %}{{ foo? }}",
        expect="hello",
        strict=True,
        error=True,
    ),
    Case(
        description="trailing question mark output",
        template="{{ bar? }}",
        expect="goodbye",
        globals={"bar?": "goodbye"},
        strict=True,
        error=False,
    ),
    # Case(
    #     description="non-trailing question mark output",
    #     template="{{ fo?o }}",
    #     expect="hello",
    #     globals={"fo?o": "hello"},
    #     strict=True,
    #     error=True,
    # ),
    Case(
        description="hyphen in for loop target",
        template="{% for x in f-oo %}{{ x }}{% endfor %}",
        expect="123",
        globals={"f-oo": [1, 2, 3]},
        strict=True,
    ),
    Case(
        description="leading hyphen in for loop target",
        template="{% for x in -foo %}{{ x }}{% endfor %}",
        expect="123",
        globals={"-foo": [1, 2, 3]},
        strict=True,
        error=True,
    ),
    Case(
        description="hyphen in for loop variable",
        template="{% for x-y in foo %}{{ x-y }}{% endfor %}",
        expect="123",
        globals={"foo": [1, 2, 3]},
        strict=True,
    ),
    Case(
        description="trailing question mark in for loop target",
        template="{% for x in foo? %}{{ x }}{% endfor %}",
        expect="123",
        globals={"foo?": [1, 2, 3]},
        strict=True,
    ),
    Case(
        description="trailing question mark in for loop variable",
        template="{% for x? in foo %}{{ x? }}{% endfor %}",
        expect="123",
        globals={"foo": [1, 2, 3]},
        strict=True,
    ),
    Case(
        description="increment with a hyphen",
        template="{% increment f-oo %}{% increment f-oo %}",
        expect="01",
        strict=True,
    ),
    # Case(
    #     description="increment with a leading hyphen",
    #     template="{% increment -foo %}{% increment -foo %}",
    #     expect="01",
    #     strict=True,
    # ),
    Case(
        description="decrement with a hyphen",
        template="{% decrement f-oo %}{% decrement f-oo %}",
        expect="-1-2",
        strict=True,
    ),
    # Case(
    #     description="decrement with a leading hyphen",
    #     template="{% decrement -foo %}{% decrement -foo %}",
    #     expect="-1-2",
    #     strict=True,
    # ),
    Case(
        description="capture ascii lowercase",
        template="{% capture foo %}hello{% endcapture %}{{ foo }}",
        expect="hello",
        strict=True,
    ),
    Case(
        description="capture ascii uppercase",
        template="{% capture FOO %}hello{% endcapture %}{{ FOO }}",
        expect="hello",
        strict=True,
    ),
    Case(
        description="capture digits",
        template="{% capture foo1 %}hello{% endcapture %}{{ foo1 }}",
        expect="hello",
        strict=True,
    ),
    Case(
        description="capture only digits",
        template="{% capture 123 %}hello{% endcapture %}{{ 123 }}",
        expect="123",
        strict=True,
    ),
    # Case(
    #     description="capture leading digits",
    #     template="{% capture 1foo %}hello {{ 2bar }}{% endcapture %}{{ 1foo }}",
    #     expect="hello goodbye",
    #     globals={"2bar": "goodbye"},
    #     strict=True,
    #     error=True,
    # ),
    Case(
        description="capture hyphens",
        template="{% capture foo-a %}hello {{ bar-b }}{% endcapture %}{{ foo-a }}",
        expect="hello goodbye",
        globals={"bar-b": "goodbye"},
        strict=True,
    ),
    Case(
        description="capture leading hyphen",
        template="{% capture -foo %}hello {{ -bar }}{% endcapture %}{{ -foo }}",
        expect="hello goodbye",
        globals={"-bar": "goodbye"},
        strict=True,
        error=True,
    ),
    Case(
        description="capture underscore",
        template="{% capture foo_a %}hello {{ bar_b }}{% endcapture %}{{ foo_a }}",
        expect="hello goodbye",
        globals={"bar_b": "goodbye"},
        strict=True,
    ),
    Case(
        description="capture only underscore",
        template="{% capture _ %}hello {{ __ }}{% endcapture %}{{ _ }}",
        expect="hello goodbye",
        globals={"__": "goodbye"},
        strict=True,
    ),
    Case(
        description="capture leading underscore",
        template="{% capture _foo %}hello {{ _bar }}{% endcapture %}{{ _foo }}",
        expect="hello goodbye",
        globals={"_bar": "goodbye"},
        strict=True,
    ),
    # Case(
    #     description="capture trailing question mark",
    #     template="{% capture foo? %}hello{% endcapture %}{{ foo? }}",
    #     expect="",
    #     strict=True,
    # ),
    Case(
        description="at sign",
        template="{{ @foo }}",
        expect="hello",
        globals={"@foo": "hello"},
        strict=True,
        error=True,
    ),
    # Case(
    #     description="assign at sign",
    #     template="{% assign @foo = 'hello' %}",
    #     expect="",
    #     strict=True,
    # ),
]



"""Golden tests cases for testing liquid's built-in `divided_by` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="integer value and integer arg",
        template=r"{{ 10 | divided_by: 2 }}",
        expect="5",
    ),
    Case(
        description="integer value and float arg",
        template=r"{{ 10 | divided_by: 2.0 }}",
        expect="5.0",
    ),
    Case(
        description="integer division",
        template=r"{{ 9 | divided_by: 2 }}",
        expect="4",
    ),
    Case(
        description="float value and integer arg",
        template=r"{{ 9.0 | divided_by: 2 }}",
        expect="4.5",
    ),
    Case(
        description="float division",
        template=r"{{ 20 | divided_by: 7.0 }}",
        expect="2.857142857142857",
    ),
    Case(
        description="string value and argument",
        template=r'{{ "10" | divided_by: "2" }}',
        expect="5",
    ),
    Case(
        description="string not a number",
        template=r'{{ "foo" | divided_by: "2" }}',
        expect="0",
    ),
    Case(
        description="arg string not a number",
        template=r'{{ "10" | divided_by: "foo" }}',
        expect="",
        error=True,
    ),
    Case(
        description="too many args",
        template=r"{{ 5 | divided_by: 1, '5' }}",
        expect="",
        error=True,
    ),
    Case(
        description="not a string, int or float",
        template=r"{{ a | divided_by: 1 }}",
        expect="0",
        globals={"a": {}},
    ),
    Case(
        description="undefined left value",
        template=r"{{ nosuchthing | divided_by: 2 }}",
        expect="0",
    ),
    Case(
        description="undefined argument",
        template=r"{{ 10 | divided_by: nosuchthing }}",
        expect="",
        error=True,
    ),
    Case(
        description="divied by zero",
        template=r"{{ 10 | divided_by: 0 }}",
        expect="",
        error=True,
    ),
    Case(
        description="zero divided by float",
        template=r"{{ 0 | divided_by: 1.1 }}",
        expect="0.0",
    ),
    Case(
        description="zero divided by integer",
        template=r"{{ 0 | divided_by: 1 }}",
        expect="0",
    ),
    Case(
        description="issue",
        template=r"{{ 5 | divided_by: 3 }}",
        expect="1",
    ),
    Case(
        description="render",
        template=r"{{ 5.0 }} {{ 5 }}",
        expect="5.0 5",
    ),
    Case(
        description="left value is an empty string",
        template=r"{{ '' | divided_by: 2 }}",
        expect="0",
    ),
]



"""Parse tree node and tag definition for the built in "render" tag."""
import sys

from typing import List
from typing import Optional
from typing import Dict
from typing import TextIO
from typing import Tuple

from liquid.ast import ChildNode
from liquid.ast import Node

from liquid.builtin.drops import IterableDrop
from liquid.builtin.tags.for_tag import ForLoop
from liquid.builtin.tags.include_tag import TAG_INCLUDE

from liquid.context import Context
from liquid.context import ReadOnlyChainMap

from liquid.exceptions import LiquidSyntaxError

from liquid.expression import Expression, Literal
from liquid.expression import Identifier

from liquid.parse import expect
from liquid.stream import TokenStream
from liquid.tag import Tag

from liquid.token import Token
from liquid.token import TOKEN_EXPRESSION
from liquid.token import TOKEN_IDENTIFIER
from liquid.token import TOKEN_WITH
from liquid.token import TOKEN_FOR
from liquid.token import TOKEN_AS
from liquid.token import TOKEN_COMMA
from liquid.token import TOKEN_COLON
from liquid.token import TOKEN_EOF
from liquid.token import TOKEN_STRING

from liquid.expressions import TokenStream as ExprTokenStream
from liquid.expressions.common import parse_string_literal
from liquid.expressions.common import parse_unchained_identifier
from liquid.expressions.common import parse_identifier
from liquid.expressions.include.lex import tokenize
from liquid.expressions.filtered.parse import parse_obj


TAG_RENDER = sys.intern("render")


class RenderNode(Node):
    """Parse tree node for the built-in "render" tag."""

    __slots__ = ("tok", "name", "var", "loop", "alias", "args")
    tag = TAG_RENDER

    def __init__(
        self,
        tok: Token,
        name: Expression,
        var: Optional[Expression] = None,
        loop: bool = False,
        alias: Optional[str] = None,
        args: Optional[Dict[str, Expression]] = None,
    ):
        self.tok = tok
        self.name = name
        self.var = var
        self.loop = loop
        self.alias = alias
        self.args = args or {}

    def __str__(self) -> str:
        buf = [f"{self.name}"]

        if self.var:
            buf.append(f" with {self.var}")

        if self.alias:
            buf.append(f" as {self.alias}")

        if self.args:
            buf.append(", ")

        args = (f"{key}={val}" for key, val in self.args.items())
        buf.append(", ".join(args))

        return f"{self.tag}({''.join(buf)})"

    def __repr__(self) -> str:
        return f"RenderNode(tok={self.tok!r}, name={self.name})"  # pragma: no cover

    def render_to_output(self, context: Context, buffer: TextIO) -> Optional[bool]:
        path = self.name.evaluate(context)
        assert isinstance(path, str)
        template = context.get_template_with_context(path, tag=self.tag)

        # Evaluate keyword arguments once. Unlike 'include', 'render' can not
        # mutate variables in the outer scope, so there's no need to re-evaluate
        # arguments for each loop (if any).
        args = {k: v.evaluate(context) for k, v in self.args.items()}

        # We're using a chain map here in case we need to push a forloop drop into
        # it. As drops are read only, the built-in collections.ChainMap will not do.
        namespace = ReadOnlyChainMap(args)

        # New context with globals and filters from the parent, plus the read only
        # namespace containing render arguments and bound variable.
        ctx = context.copy(
            namespace,
            disabled_tags=[TAG_INCLUDE],
            carry_loop_iterations=True,
        )

        # Optionally bind a variable to the render namespace.
        if self.var is not None:
            val = self.var.evaluate(context)
            key = self.alias or template.name.split(".")[0]

            # If the variable is array-like, render the template once for each item.
            # `self.loop` being True indicates the render expression used "for" not
            # "with". This distinction is not made when using the 'include' tag.
            if self.loop and isinstance(val, (tuple, list, IterableDrop)):
                ctx.raise_for_loop_limit(len(val))
                forloop = ForLoop(
                    name=key,
                    it=iter(val),
                    length=len(val),
                    parentloop=context.env.undefined("parentloop"),
                )

                args["forloop"] = forloop
                args[key] = None

                for itm in forloop:
                    args[key] = itm
                    template.render_with_context(
                        ctx, buffer, partial=True, block_scope=True
                    )
            else:
                # The bound variable is not array-like, shove it into the namespace
                # via args.
                args[key] = val
                template.render_with_context(
                    ctx, buffer, partial=True, block_scope=True
                )
        else:
            template.render_with_context(ctx, buffer, partial=True, block_scope=True)

        return True

    async def render_to_output_async(
        self, context: Context, buffer: TextIO
    ) -> Optional[bool]:
        """An awaitable version of `render_to_output` that loads templates
        asynchronously."""
        path = await self.name.evaluate_async(context)
        assert isinstance(path, str)
        template = await context.get_template_with_context_async(path, tag=self.tag)

        # Evaluate keyword arguments once. Unlike 'include', 'render' can not
        # mutate variables in the outer scope, so there's no need to re-evaluate
        # arguments for each loop (if any).
        args = {k: await v.evaluate_async(context) for k, v in self.args.items()}

        # We're using a chain map here in case we need to push a forloop drop into
        # it. As drops are read only, the built-in collections.ChainMap will not do.
        namespace = ReadOnlyChainMap(args)

        # New context with globals and filters from the parent, plus the read only
        # namespace containing render arguments and bound variable.
        ctx = context.copy(
            namespace,
            disabled_tags=[TAG_INCLUDE],
            carry_loop_iterations=True,
        )

        # Optionally bind a variable to the render namespace.
        if self.var is not None:
            val = await self.var.evaluate_async(context)
            key = self.alias or template.name.split(".")[0]

            # If the variable is array-like, render the template once for each item.
            # `self.loop` being True indicates the render expression used "for" not
            # "with". This distinction is not made when using the 'include' tag.
            if self.loop and isinstance(val, (tuple, list, IterableDrop)):
                ctx.raise_for_loop_limit(len(val))
                forloop = ForLoop(
                    name=key,
                    it=iter(val),
                    length=len(val),
                    parentloop=context.env.undefined("parentloop"),
                )

                args["forloop"] = forloop
                args[key] = None

                for itm in forloop:
                    args[key] = itm
                    await template.render_with_context_async(
                        ctx, buffer, partial=True, block_scope=True
                    )
            else:
                # The bound variable is not array-like, shove it into the namespace
                # via args.
                args[key] = val
                await template.render_with_context_async(
                    ctx, buffer, partial=True, block_scope=True
                )
        else:
            await template.render_with_context_async(
                ctx, buffer, partial=True, block_scope=True
            )

        return True

    def children(self) -> List[ChildNode]:
        block_scope: List[str] = list(self.args.keys())
        _children = [
            ChildNode(
                linenum=self.tok.linenum,
                node=None,
                expression=self.name,
                block_scope=block_scope,
                load_mode="render",
                load_context={"tag": "render"},
            )
        ]
        if self.var:
            if self.alias:
                block_scope.append(self.alias)
            elif isinstance(self.name, Literal):
                block_scope.append(str(self.name.value).split(".", 1)[0])
            _children.append(
                ChildNode(
                    linenum=self.tok.linenum,
                    expression=self.var,
                )
            )
        for expr in self.args.values():
            _children.append(ChildNode(linenum=self.tok.linenum, expression=expr))
        return _children


BIND_TAGS = frozenset((TOKEN_WITH, TOKEN_FOR))


class RenderTag(Tag):
    """The built-in "render" tag."""

    name = TAG_RENDER
    block = False
    node_class = RenderNode

    def parse(self, stream: TokenStream) -> Node:
        tok = next(stream)
        expect(stream, TOKEN_EXPRESSION)
        expr_stream = ExprTokenStream(tokenize(stream.current.value))

        # Need a string. 'render' does not accept identifiers that resolve to a string.
        # This is the name of the template to be included.
        expr_stream.expect(TOKEN_STRING)
        name = parse_string_literal(expr_stream)
        expr_stream.next_token()

        alias: Optional[str] = None
        identifier: Optional[Identifier] = None
        loop: bool = False

        # Optionally bind a variable to the included template context
        if expr_stream.current[1] in BIND_TAGS:
            loop = expr_stream.current[1] == TOKEN_FOR
            expr_stream.next_token()  # Eat 'with' or 'for'
            expr_stream.expect(TOKEN_IDENTIFIER)
            identifier = parse_identifier(expr_stream)
            expr_stream.next_token()

            # The bound variable will take the name of the template by default,
            # or an alias if an identifier follows the "as" keyword.
            if expr_stream.current[1] == TOKEN_AS:
                expr_stream.next_token()  # Eat 'as'
                expr_stream.expect(TOKEN_IDENTIFIER)
                alias = str(parse_unchained_identifier(expr_stream))
                expr_stream.next_token()

        # Zero or more keyword arguments
        args = {}

        # The first keyword argument might follow immediately or after a comma.
        if expr_stream.current[1] == TOKEN_IDENTIFIER:
            key, val = parse_argument(expr_stream)
            args[key] = val

        while expr_stream.current[1] != TOKEN_EOF:
            if expr_stream.current[1] == TOKEN_COMMA:
                expr_stream.next_token()  # Eat comma
                key, val = parse_argument(expr_stream)
                args[key] = val
            else:
                typ = expr_stream.current[1]
                raise LiquidSyntaxError(
                    f"expected a comma separated list of arguments, found {typ}",
                    linenum=tok.linenum,
                )

        return self.node_class(
            tok,
            name=name,
            var=identifier,
            loop=loop,
            alias=alias,
            args=args,
        )


def parse_argument(stream: ExprTokenStream) -> Tuple[str, Expression]:
    """Return the next key/value pair from the stream where key and value
    are separated by a colon."""
    key = str(parse_unchained_identifier(stream))
    stream.next_token()
    stream.expect(TOKEN_COLON)
    stream.next_token()  # Eat colon
    val = parse_obj(stream)
    stream.next_token()
    return key, val



"""Tag and node definition for the built-in "case" tag."""
from __future__ import annotations
import sys

from typing import List
from typing import Optional
from typing import TextIO
from typing import TYPE_CHECKING

from liquid.token import Token
from liquid.token import TOKEN_EOF
from liquid.token import TOKEN_EXPRESSION
from liquid.token import TOKEN_TAG

from liquid.parse import get_parser
from liquid.parse import expect

from liquid import ast
from liquid.tag import Tag
from liquid.context import Context
from liquid.stream import TokenStream

if TYPE_CHECKING:
    from liquid import Environment
    from liquid.expression import Expression

TAG_CASE = sys.intern("case")
TAG_ENDCASE = sys.intern("endcase")
TAG_WHEN = sys.intern("when")
TAG_ELSE = sys.intern("else")

ENDWHENBLOCK = frozenset((TAG_ENDCASE, TAG_WHEN, TAG_ELSE, TOKEN_EOF))
ENDCASEBLOCK = frozenset((TAG_ENDCASE,))


class CaseNode(ast.Node):
    """Parse tree node for the built-in "case" tag."""

    __slots__ = ("tok", "whens", "default", "forced_output")

    def __init__(
        self,
        tok: Token,
        whens: List[ast.ConditionalBlockNode],
        default: Optional[ast.BlockNode] = None,
    ):
        self.tok = tok
        self.whens = whens
        self.default = default

        self.forced_output = any(
            n.forced_output for n in (*self.whens, self.default) if n
        )

    def __str__(self) -> str:
        if not self.whens:
            buf = ["if (False) { }"]
        else:
            buf = [f"if {self.whens[0]}"]

        if len(self.whens) > 1:
            for when in self.whens[1:]:
                buf.append(f"elsif {when}")

        if self.default:
            buf.append(f"else {{ {self.default} }}")

        return " ".join(buf)

    def render_to_output(self, context: Context, buffer: TextIO) -> Optional[bool]:
        buf = context.get_buffer(buffer)
        rendered: Optional[bool] = False

        for when in self.whens:
            if when.render(context, buf):
                rendered = True

        if not rendered and self.default:
            rendered = self.default.render(context, buf)

        val = buf.getvalue()
        if self.forced_output or not val.isspace():
            buffer.write(val)

        return rendered

    async def render_to_output_async(
        self, context: Context, buffer: TextIO
    ) -> Optional[bool]:
        buf = context.get_buffer(buffer)
        rendered: Optional[bool] = False

        for when in self.whens:
            if await when.render_async(context, buf):
                rendered = True

        if not rendered and self.default:
            rendered = await self.default.render_async(context, buf)

        val = buf.getvalue()
        if self.forced_output or not val.isspace():
            buffer.write(val)

        return rendered

    def children(self) -> List[ast.ChildNode]:
        _children = [
            ast.ChildNode(
                linenum=alt.tok.linenum,
                node=alt.block,
                expression=alt.condition,
            )
            for alt in self.whens
        ]
        if self.default:
            _children.append(
                ast.ChildNode(
                    linenum=self.default.tok.linenum,
                    node=self.default,
                    expression=None,
                )
            )
        return _children


class CaseTag(Tag):
    """The built-in cycle tag."""

    name = TAG_CASE
    end = TAG_ENDCASE

    def __init__(self, env: Environment):
        super().__init__(env)
        self.parser = get_parser(self.env)

    def parse_expression(self, case: str, obj: str, stream: TokenStream) -> Expression:
        """Parse a boolean expression from a stream of tokens."""
        expect(stream, TOKEN_EXPRESSION)
        return self.env.parse_boolean_expression_value(f"{case} == {obj}")

    def parse(self, stream: TokenStream) -> ast.Node:
        expect(stream, TOKEN_TAG, value=TAG_CASE)
        tok = stream.current
        stream.next_token()

        expect(stream, TOKEN_EXPRESSION)
        case = stream.current.value
        stream.next_token()

        # Eat whitespace or junk between `case` and when/else/endcase
        while (
            stream.current.type != TOKEN_TAG
            and stream.current.value not in ENDWHENBLOCK
        ):
            stream.next_token()

        whens: List[ast.ConditionalBlockNode] = []

        while stream.current.istag(TAG_WHEN):
            when_tok = stream.current
            stream.next_token()  # Eat WHEN

            # One conditional block for every object in a comma separated list.
            when_exprs = [
                self.parse_expression(case, obj, stream)
                for obj in stream.current.value.split(",")
            ]

            stream.next_token()
            when_block = self.parser.parse_block(stream, ENDWHENBLOCK)

            whens.extend(
                ast.ConditionalBlockNode(
                    tok=when_tok,
                    condition=expr,
                    block=when_block,
                )
                for expr in when_exprs
            )

        default: Optional[ast.BlockNode] = None

        if stream.current.istag(TAG_ELSE):
            stream.next_token()
            default = self.parser.parse_block(stream, ENDCASEBLOCK)

        expect(stream, TOKEN_TAG, value=TAG_ENDCASE)
        return CaseNode(tok, whens=whens, default=default)



"""Golden tests cases for testing liquid's built-in `prepend` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="concat",
        template=r'{{ "hello" | prepend: "there" }}',
        expect="therehello",
    ),
    Case(
        description="not a string",
        template=r"{{ 5 | prepend: 'there' }}",
        expect="there5",
    ),
    Case(
        description="argument not a string",
        template=r'{{ "hello" | prepend: 5 }}',
        expect="5hello",
    ),
    Case(
        description="missing argument",
        template=r'{{ "hello" | prepend }}',
        expect="",
        error=True,
    ),
    Case(
        description="too many arguments",
        template=r'{{ "hello" | prepend: "how", "are", "you" }}',
        expect="",
        error=True,
    ),
    Case(
        description="undefined left value",
        template=r'{{ nosuchthing | prepend: "hi" }}',
        expect="hi",
    ),
    Case(
        description="undefined argument",
        template=r'{{ "hi" | prepend: nosuchthing }}',
        expect="hi",
    ),
]



"""Golden tests cases for testing liquid's built-in `first` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="range literal first filter left value",
        template=r"{{ (1..3) | first }}",
        expect="1",
    ),
    Case(
        description="array of strings",
        template=r"{{ arr | first }}",
        expect="a",
        globals={"arr": ["a", "b"]},
    ),
    Case(
        description="array of things",
        template=r"{{ arr | first }}",
        expect="a",
        globals={"arr": ["a", "b", 1, [], {}]},
    ),
    Case(
        description="empty left value",
        template=r"{{ arr | first }}",
        expect="",
        globals={"arr": []},
    ),
    Case(
        description="left value is not an array",
        template=r"{{ arr | first }}",
        expect="",
        globals={"arr": 12},
    ),
    Case(
        description="left value is undefined",
        template=r"{{ nosuchthing | first }}",
        expect="",
    ),
    Case(
        description="first of a string",
        template=r"{{ 'hello' | first }}",
        expect="",
    ),
    Case(
        description="first of a hash",
        template=r"{% assign x = a | first %}({{ x[0] }},{{ x[1] }})",
        expect="(b,1)",
        globals={"a": {"b": 1, "c": 2}},
    ),
]



"""A mock JSON encoding filter."""

import json

from liquid.filter import liquid_filter
from liquid.exceptions import FilterArgumentError


@liquid_filter
def json_(mapping):
    """A filter function that dumps a dictionary in JSON."""
    if not isinstance(mapping, dict):
        raise FilterArgumentError(
            f"json expected an object, found {type(mapping).__name__}"
        )
    return json.dumps({k: v for k, v in mapping.items() if k != "collections"})



"""A simple LRU Cache implementation.

Coppied from https://github.com/pallets/jinja/blob/master/src/jinja2/utils.py
BSD-3-Clause License

Copyright 2007 Pallets

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from collections import abc, deque
from threading import Lock


# pylint: disable=too-many-instance-attributes,invalid-name,protected-access,consider-using-with
class LRUCache(abc.MutableMapping):
    """A simple LRU Cache implementation."""

    # this is fast for small capacities (something below 1000) but doesn't
    # scale.  But as long as it's only used as storage for templates this
    # won't do any harm.

    def __init__(self, capacity: int):
        self.capacity = capacity
        self._mapping = {}
        self._queue = deque()
        self._postinit()

    def _postinit(self):
        # alias all queue methods for faster lookup
        self._popleft = self._queue.popleft
        self._pop = self._queue.pop
        self._remove = self._queue.remove
        self._wlock = Lock()
        self._append = self._queue.append

    def __getstate__(self):  # pragma: no cover
        return {
            "capacity": self.capacity,
            "_mapping": self._mapping,
            "_queue": self._queue,
        }

    def __setstate__(self, d):  # pragma: no cover
        self.__dict__.update(d)
        self._postinit()

    def __getnewargs__(self):  # pragma: no cover
        return (self.capacity,)

    def copy(self):
        """Return a shallow copy of the instance."""
        rv = self.__class__(self.capacity)
        rv._mapping.update(self._mapping)
        rv._queue.extend(self._queue)
        return rv

    def get(self, key: str, default: object = None):
        """Return an item from the cache dict or `default`"""
        try:
            return self[key]
        except KeyError:
            return default

    def setdefault(self, key, default=None):
        """Set `default` if the key is not in the cache otherwise
        leave unchanged. Return the value of this key.
        """
        try:
            return self[key]
        except KeyError:
            self[key] = default
            return default

    def clear(self):
        """Clear the cache."""
        self._wlock.acquire()
        try:
            self._mapping.clear()
            self._queue.clear()
        finally:
            self._wlock.release()

    def __contains__(self, key):
        """Check if a key exists in this cache."""
        return key in self._mapping

    def __len__(self):
        """Return the current size of the cache."""
        return len(self._mapping)

    def __repr__(self):  # pragma: no cover
        return f"<{self.__class__.__name__} {self._mapping!r}>"

    def __getitem__(self, key):
        """Get an item from the cache. Moves the item up so that it has the
        highest priority then.
        Raise a `KeyError` if it does not exist.
        """
        self._wlock.acquire()
        try:
            rv = self._mapping[key]
            if self._queue[-1] != key:
                try:
                    self._remove(key)
                except ValueError:  # pragma: no cover
                    # if something removed the key from the container
                    # when we read, ignore the ValueError that we would
                    # get otherwise.
                    pass
                self._append(key)
            return rv
        finally:
            self._wlock.release()

    def __setitem__(self, key, value):
        """Sets the value for an item. Moves the item up so that it
        has the highest priority then.
        """
        self._wlock.acquire()
        try:
            if key in self._mapping:
                self._remove(key)
            elif len(self._mapping) == self.capacity:
                del self._mapping[self._popleft()]
            self._append(key)
            self._mapping[key] = value
        finally:
            self._wlock.release()

    def __delitem__(self, key):
        """Remove an item from the cache dict.
        Raise a `KeyError` if it does not exist.
        """
        self._wlock.acquire()
        try:
            del self._mapping[key]
            try:
                self._remove(key)
            except ValueError:  # pragma: no cover
                pass
        finally:
            self._wlock.release()

    def items(self):
        """Return a list of items."""
        result = [(key, self._mapping[key]) for key in list(self._queue)]
        result.reverse()
        return result

    def values(self):
        """Return a list of all values."""
        return [x[1] for x in self.items()]

    def keys(self):
        """Return a list of all keys ordered by most recent usage."""
        return list(self)

    def __iter__(self):
        return reversed(tuple(self._queue))

    def __reversed__(self):
        """Iterate over the keys in the cache dict, oldest items
        coming first.
        """
        return iter(tuple(self._queue))

    __copy__ = copy



"""Functions for parsing non-standard conditional expressions."""
from functools import partial

from typing import Dict
from typing import List
from typing import Iterator
from typing import Iterable
from typing import Optional
from typing import Tuple

from liquid.expression import BooleanExpression
from liquid.expression import Expression
from liquid.expression import FALSE
from liquid.expression import Filter
from liquid.expression import ConditionalExpression
from liquid.expression import FilteredExpression

from liquid.expressions.boolean.parse import parse_obj as parse_boolean_obj
from liquid.expressions.boolean.parse import (
    parse_obj_with_parens as parse_boolean_obj_with_parens,
)
from liquid.expressions.common import Token
from liquid.expressions.conditional.lex import tokenize
from liquid.expressions.conditional.lex import tokenize_with_parens
from liquid.expressions.filtered.parse import (
    parse_from_tokens as parse_standard_filtered,
)
from liquid.expressions.filtered.parse import parse_obj
from liquid.expressions.stream import TokenStream

from liquid.exceptions import LiquidSyntaxError

from liquid.token import TOKEN_EOF
from liquid.token import TOKEN_PIPE
from liquid.token import TOKEN_DPIPE
from liquid.token import TOKEN_COMMA
from liquid.token import TOKEN_IDENTIFIER
from liquid.token import TOKEN_COLON
from liquid.token import TOKEN_IF
from liquid.token import TOKEN_ELSE


def split_at_pipe(tokens: Iterable[Token]) -> Iterator[List[Token]]:
    """Split tokens on into lists, using TOKEN_PIPE as the delimiter."""
    buf: List[Token] = []
    for token in tokens:
        if token[1] == TOKEN_PIPE:
            yield buf
            buf = []
        else:
            buf.append(token)
    yield buf


def _split_at_first(
    tokens: Iterator[Token],
    _type: str,
) -> Tuple[List[Token], Optional[Iterator[Token]]]:
    buf: List[Token] = []
    for token in tokens:
        if token[1] == _type:
            return (buf, tokens)
        buf.append(token)
    return (buf, None)


split_at_first_pipe = partial(_split_at_first, _type=TOKEN_PIPE)
split_at_first_dpipe = partial(_split_at_first, _type=TOKEN_DPIPE)
split_at_first_if = partial(_split_at_first, _type=TOKEN_IF)
split_at_first_else = partial(_split_at_first, _type=TOKEN_ELSE)


def _parse_filter(tokens: List[Token], linenum: int) -> Filter:
    if not tokens:
        raise LiquidSyntaxError(
            "unexpected pipe or missing filter name", linenum=linenum
        )

    stream = TokenStream(iter(tokens))
    stream.expect(TOKEN_IDENTIFIER)
    name = stream.current[2]

    next(stream)
    # Shortcut for filters with no arguments.
    if stream.current[1] == TOKEN_EOF:
        return Filter(name, [])

    # Eat colon
    stream.expect(TOKEN_COLON)
    next(stream)

    args: List[Expression] = []
    kwargs: Dict[str, Expression] = {}

    while stream.current[1] != TOKEN_EOF:
        if stream.peek[1] == TOKEN_COLON:
            # A keyword argument
            stream.expect(TOKEN_IDENTIFIER)
            key = next(stream)[2]
            # Eat colon
            next(stream)
            kwargs[key] = parse_obj(stream)
        else:
            # A positional argument
            args.append(parse_obj(stream))

        # Eat comma
        next(stream)
        if stream.current[1] != TOKEN_EOF:
            stream.expect(TOKEN_COMMA)
            next(stream)

    return Filter(name, args, kwargs)


def parse(expr: str, linenum: int = 1) -> FilteredExpression:
    """Parse a conditional expression string."""
    tokens = tokenize(expr, linenum)
    standard_tokens, _conditional_tokens = split_at_first_if(tokens)

    # This expression includes filters.
    _expr = parse_standard_filtered(iter(standard_tokens), linenum)

    if not _conditional_tokens:
        # A standard filtered expression
        return _expr

    conditional_tokens, _filter_tokens = split_at_first_dpipe(_conditional_tokens)
    conditional_tokens, _alternative_tokens = split_at_first_else(
        iter(conditional_tokens)
    )

    if conditional_tokens:
        condition = BooleanExpression(
            parse_boolean_obj(TokenStream(iter(conditional_tokens)), linenum)
        )
    else:
        # A missing condition (an `if` with nothing after it).
        condition = BooleanExpression(FALSE)

    if _alternative_tokens:
        # Handle TOKEN_ELSE followed by nothing.
        alternative_tokens: List[Token] = list(_alternative_tokens)

        if alternative_tokens:
            alternative: Optional[Expression] = parse_standard_filtered(
                iter(alternative_tokens), linenum
            )
        else:
            alternative = None
    else:
        alternative = None

    if _filter_tokens:
        tail_filters = [
            _parse_filter(_tokens, linenum)
            for _tokens in split_at_pipe(iter(_filter_tokens))
        ]
    else:
        tail_filters = []

    return ConditionalExpression(_expr, tail_filters, condition, alternative)


def parse_with_parens(expr: str, linenum: int = 1) -> FilteredExpression:
    """Parse a conditional expression string that supports logical `not` and
    grouping terms with parentheses."""
    tokens = tokenize_with_parens(expr, linenum)
    standard_tokens, _conditional_tokens = split_at_first_if(tokens)

    # This expression includes filters.
    _expr = parse_standard_filtered(iter(standard_tokens), linenum)

    if not _conditional_tokens:
        # A standard filtered expression
        return _expr

    conditional_tokens, _filter_tokens = split_at_first_dpipe(_conditional_tokens)
    conditional_tokens, _alternative_tokens = split_at_first_else(
        iter(conditional_tokens)
    )

    if conditional_tokens:
        condition = BooleanExpression(
            parse_boolean_obj_with_parens(
                TokenStream(iter(conditional_tokens)), linenum
            )
        )
    else:
        # A missing condition (an `if` with nothing after it).
        condition = BooleanExpression(FALSE)

    if _alternative_tokens:
        # Handle TOKEN_ELSE followed by nothing.
        alternative_tokens: List[Token] = list(_alternative_tokens)

        if alternative_tokens:
            alternative: Optional[Expression] = parse_standard_filtered(
                iter(alternative_tokens), linenum
            )
        else:
            alternative = None
    else:
        alternative = None

    if _filter_tokens:
        tail_filters = [
            _parse_filter(_tokens, linenum)
            for _tokens in split_at_pipe(iter(_filter_tokens))
        ]
    else:
        tail_filters = []

    return ConditionalExpression(_expr, tail_filters, condition, alternative)



"""Golden tests cases for testing text that is not liquid."""

from liquid.golden.case import Case

cases = [
    Case(
        description="plain text gets passed through unchanged",
        template="a literal string",
        expect="a literal string",
    ),
    Case(
        description="css text gets passed through unchanged",
        template=" div { font-weight: bold; } ",
        expect=" div { font-weight: bold; } ",
    ),
]



"""Test filter decorators and helpers."""
# pylint: disable=missing-class-docstring
from unittest import TestCase

from liquid import Environment
from liquid import Context

from liquid.filter import liquid_filter
from liquid.filter import with_context
from liquid.filter import num_arg
from liquid.filter import int_arg

from liquid.exceptions import FilterArgumentError


@liquid_filter
@with_context
def some_filter(val, arg, *, context: Context) -> str:
    """Contrived filter function making use of `with_context`."""
    return val + context.resolve(arg)


class WithContextTestCase(TestCase):
    def test_filter_with_context(self):
        """Test that we can pass the active context to a filter function."""
        env = Environment(strict_filters=True)
        env.add_filter("some", some_filter)

        template = env.from_string(r"{{ 'Hello, ' | some: 'you' }}")
        result = template.render(you="World")

        self.assertEqual(result, "Hello, World")


class IntArgTestCase(TestCase):
    def test_int_arg_fail(self):
        """Test that a suitable exception is raised if we can't cast to an int."""
        with self.assertRaises(FilterArgumentError):
            int_arg("foo")

    def test_int_arg_with_default(self):
        """Test that a default is returned if we can't cast to an int."""
        self.assertEqual(int_arg("foo", 99), 99)


class NumArgTestCase(TestCase):
    def test_num_arg_fail_string(self):
        """Test that a suitable exception is raised if we can't cast to a number."""
        with self.assertRaises(FilterArgumentError):
            num_arg("foo")

    def test_num_arg_with_default(self):
        """Test that a default is returned if we can't cast to a number."""
        self.assertEqual(num_arg("foo", 99.8), 99.8)

    def test_num_arg_no_default(self):
        """Test that an exception is raised if a default is not given."""
        with self.assertRaises(FilterArgumentError):
            num_arg(object())



# flake8: noqa
# pylint: disable=useless-import-alias,missing-module-docstring

__version__ = "1.5.0"

try:
    from markupsafe import escape as escape
    from markupsafe import Markup as Markup
    from markupsafe import soft_str
except ImportError:
    from liquid.exceptions import escape as escape  # type: ignore
    from liquid.exceptions import Markup as Markup  # type: ignore

    # pylint: disable=invalid-name
    soft_str = str  # type: ignore

from .mode import Mode as Mode
from .token import Token
from .expression import Expression

from .loaders import FileSystemLoader as FileSystemLoader
from .loaders import DictLoader as DictLoader
from .loaders import ChoiceLoader as ChoiceLoader

from .context import Context as Context
from .context import Undefined as Undefined
from .context import DebugUndefined as DebugUndefined
from .context import StrictUndefined as StrictUndefined
from .context import StrictDefaultUndefined as StrictDefaultUndefined
from .context import is_undefined as is_undefined

from .environment import Environment as Environment
from .environment import Template as Template



"""Golden tests cases for testing liquid's `if` tag."""

from liquid.golden.case import Case

cases = [
    Case(
        description="condition with literal consequence",
        template=r"{% if product.title == 'foo' %}bar{% endif %}",
        expect="bar",
        globals={"product": {"title": "foo"}},
    ),
    Case(
        description="condition with literal consequence and literal alternative",
        template=r"{% if product.title == 'hello' %}bar{% else %}baz{% endif %}",
        expect="baz",
        globals={"product": {"title": "foo"}},
    ),
    Case(
        description="condition with conditional alternative",
        template=(
            r"{% if product.title == 'hello' %}"
            r"foo"
            r"{% elsif product.title == 'foo' %}"
            r"bar"
            r"{% endif %}"
        ),
        expect="bar",
        globals={"product": {"title": "foo"}},
    ),
    Case(
        description="condition with conditional alternative and final alternative",
        template=(
            r"{% if product.title == 'hello' %}"
            r"foo"
            r"{% elsif product.title == 'goodbye' %}"
            r"bar"
            r"{% else %}"
            r"baz"
            r"{% endif %}"
        ),
        expect="baz",
        globals={"product": {"title": "foo"}},
    ),
    Case(
        description="non-empty hash is truthy",
        template=r"{% if product %}bar{% else %}foo{% endif %}",
        expect="bar",
        globals={"product": {"title": "foo"}},
    ),
    Case(
        description="literal nil is falsy",
        template=r"{% if nil %}bar{% else %}foo{% endif %}",
        expect="foo",
    ),
    Case(
        description="undefined variables are falsy",
        template=r"{% if nosuchthing %}bar{% else %}foo{% endif %}",
        expect="foo",
    ),
    Case(
        description="nested condition in the consequence block",
        template=(
            r"{% if product %}"
            r"{% if title == 'Hello' %}"
            r"baz"
            r"{% endif %}"
            r"{% endif %}"
        ),
        expect="baz",
        globals={
            "product": {"title": "foo"},
            "title": "Hello",
        },
    ),
    Case(
        description="nested condition, alternative in the consequence block",
        template=(
            r"{% if product %}"
            r"{% if title == 'goodbye' %}"
            r"baz"
            r"{% else %}"
            r"hello"
            r"{% endif %}"
            r"{% endif %}"
        ),
        expect="hello",
        globals={"product": {"title": "foo"}, "title": "Hello"},
    ),
    Case(
        description="literal false condition",
        template=r"{% if false %}{% endif %}",
        expect="",
    ),
    Case(
        description="contains condition",
        template=r"{% if product.tags contains 'garden' %}baz{% endif %}",
        expect="baz",
        globals={"product": {"tags": ["sports", "garden"]}},
    ),
    Case(
        description="not equal condition",
        template=r"{% if product.title != 'foo' %}baz{% endif %}",
        expect="",
        globals={"product": {"title": "foo"}},
    ),
    Case(
        description="alternate not equal condition",
        template=r"{% if product.title <> 'foo' %}baz{% endif %}",
        expect="",
        globals={"product": {"title": "foo"}},
    ),
    Case(
        description="blocks that contain only whitespace are not rendered",
        template=r"{% if true %}  {% elsif false %} {% else %} {% endif %}",
        expect="",
    ),
    Case(
        description=(
            "blocks that contain only whitespace and comments are not rendered"
        ),
        template=(
            r"{% if true %} "
            r"{% comment %} this is blank {% endcomment %} "
            r"{% endif %}"
        ),
        expect="",
    ),
    Case(
        description="compare empty string literal to blank",
        template=r"{% if '' == blank %}is blank{% endif %}",
        expect="is blank",
        standard=False,
    ),
    Case(
        description="conditional alternative with default",
        template=(
            r"{% if false %}foo"
            r"{% elsif false %}bar"
            r"{% else %}hello"
            r"{% endif %}"
        ),
        expect="hello",
    ),
    Case(
        description="range equals range",
        template=(
            r"{% assign foo = (1..3) %}"
            r"{% if foo == (1..3) %}true"
            r"{% else %}false"
            r"{% endif %}"
        ),
        expect="true",
    ),
    Case(
        description="logical operators are right associative",
        template=(r"{% if true and false and false or true %}hello{% endif %}"),
        expect="",
    ),
    Case(
        description=("zero is not equal to false"),
        template=(r"{% if 0 == false %}Hello{% else %}Goodbye{% endif %}"),
        expect="Goodbye",
    ),
    Case(
        description=("zero is truthy"),
        template=(r"{% if 0 %}Hello{% else %}Goodbye{% endif %}"),
        expect="Hello",
    ),
    Case(
        description=("0.0 is truthy"),
        template=(r"{% if 0.0 %}Hello{% else %}Goodbye{% endif %}"),
        expect="Hello",
    ),
    Case(
        description=("one is not equal to true"),
        template=(r"{% if 1 == true %}Hello{% else %}Goodbye{% endif %}"),
        expect="Goodbye",
    ),
]
